# Copyright (c) 2015-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Compositors filling one composite with others."""

from __future__ import annotations

import logging
from typing import Optional, Sequence

import dask.array as da
import numpy as np
import xarray as xr

from satpy.dataset import combine_metadata

from .core import (
    GenericCompositor,
    SingleBandCompositor,
    add_bands,
    enhance2dataset,
)

LOG = logging.getLogger(__name__)


class FillingCompositor(GenericCompositor):
    """Make a regular RGB, filling the RGB bands with the first provided dataset's values."""

    def __call__(self, projectables, nonprojectables=None, **info):
        """Generate the composite."""
        projectables = self.match_data_arrays(projectables)
        projectables[1] = projectables[1].fillna(projectables[0])
        projectables[2] = projectables[2].fillna(projectables[0])
        projectables[3] = projectables[3].fillna(projectables[0])
        return super(FillingCompositor, self).__call__(projectables[1:], **info)


class Filler(GenericCompositor):
    """Fix holes in projectable 1 with data from projectable 2."""

    def __call__(self, projectables, nonprojectables=None, **info):
        """Generate the composite."""
        projectables = self.match_data_arrays(projectables)
        filled_projectable = projectables[0].fillna(projectables[1])
        return super(Filler, self).__call__([filled_projectable], **info)


class MultiFiller(SingleBandCompositor):
    """Fix holes in projectable 1 with data from the next projectables."""

    def __call__(self, projectables, nonprojectables=None, **info):
        """Generate the composite."""
        projectables = self.match_data_arrays(projectables)
        filled_projectable = projectables[0]
        for next_projectable in projectables[1:]:
            filled_projectable = filled_projectable.fillna(next_projectable)
        if "optional_datasets" in info.keys():
            for next_projectable in info["optional_datasets"]:
                filled_projectable = filled_projectable.fillna(next_projectable)

        return super().__call__([filled_projectable], **info)


class DayNightCompositor(GenericCompositor):
    """A compositor that blends day data with night data.

    Using the `day_night` flag it is also possible to provide only a day product
    or only a night product and mask out (make transparent) the opposite portion
    of the image (night or day). See the documentation below for more details.
    """

    def __init__(self, name, lim_low=85., lim_high=88., day_night="day_night", include_alpha=True, **kwargs):  # noqa: D417
        """Collect custom configuration values.

        Args:
            lim_low (float): lower limit of Sun zenith angle for the
                             blending of the given channels
            lim_high (float): upper limit of Sun zenith angle for the
                             blending of the given channels
            day_night (str): "day_night" means both day and night portions will be kept
                                "day_only" means only day portion will be kept
                                "night_only" means only night portion will be kept
            include_alpha (bool): This only affects the "day only" or "night only" result.
                                  True means an alpha band will be added to the output image for transparency.
                                  False means the output is a single-band image with undesired pixels being masked out
                                  (replaced with NaNs).

        """
        self.lim_low = lim_low
        self.lim_high = lim_high
        self.day_night = day_night
        self.include_alpha = include_alpha
        self._has_sza = False
        super().__init__(name, **kwargs)

    def __call__(
            self,
            datasets: Sequence[xr.DataArray],
            optional_datasets: Optional[Sequence[xr.DataArray]] = None,
            **attrs
    ) -> xr.DataArray:
        """Generate the composite."""
        datasets = self.match_data_arrays(datasets)
        # At least one composite is requested.
        foreground_data = datasets[0]
        weights = self._get_coszen_blending_weights(datasets)
        # Apply enhancements to the foreground data
        foreground_data = enhance2dataset(foreground_data)

        if "only" in self.day_night:
            fg_attrs = foreground_data.attrs.copy()
            day_data, night_data, weights = self._get_data_for_single_side_product(foreground_data, weights)
        else:
            day_data, night_data, fg_attrs = self._get_data_for_combined_product(foreground_data, datasets[1])

        # The computed coszen is for the full area, so it needs to be masked for missing and off-swath data
        if self.include_alpha and not self._has_sza:
            weights = self._mask_weights_with_data(weights, day_data, night_data)

        if "only" not in self.day_night:
            # Replace missing channel data with zeros
            day_data = zero_missing_data(day_data, night_data)
            night_data = zero_missing_data(night_data, day_data)

        data = self._weight_data(day_data, night_data, weights, fg_attrs)

        return super(DayNightCompositor, self).__call__(
            data,
            optional_datasets=optional_datasets,
            **attrs
        )

    def _get_coszen_blending_weights(
            self,
            projectables: Sequence[xr.DataArray],
    ) -> xr.DataArray:
        lim_low = float(np.cos(np.deg2rad(self.lim_low)))
        lim_high = float(np.cos(np.deg2rad(self.lim_high)))
        try:
            coszen = np.cos(np.deg2rad(projectables[2 if self.day_night == "day_night" else 1]))
            self._has_sza = True
        except IndexError:
            from satpy.modifiers.angles import get_cos_sza
            LOG.debug("Computing sun zenith angles.")
            # Get chunking that matches the data
            coszen = get_cos_sza(projectables[0])
        # Calculate blending weights
        coszen -= min(lim_high, lim_low)
        coszen /= abs(lim_low - lim_high)
        return coszen.clip(0, 1)

    def _get_data_for_single_side_product(
            self,
            foreground_data: xr.DataArray,
            weights: xr.DataArray,
    ) -> tuple[xr.DataArray, xr.DataArray, xr.DataArray]:
        # Only one portion (day or night) is selected. One composite is requested.
        # Add alpha band to single L/RGB composite to make the masked-out portion transparent when needed
        # L -> LA
        # RGB -> RGBA
        if self.include_alpha:
            foreground_data = add_alpha_bands(foreground_data)
        else:
            weights = self._mask_weights(weights)

        day_data, night_data = self._get_day_night_data_for_single_side_product(foreground_data)
        return day_data, night_data, weights

    def _mask_weights(self, weights):
        if "day" in self.day_night:
            return weights.where(weights != 0, np.nan)
        return weights.where(weights != 1, np.nan)

    def _get_day_night_data_for_single_side_product(self, foreground_data):
        if "day" in self.day_night:
            return foreground_data, foreground_data.dtype.type(0)
        return foreground_data.dtype.type(0), foreground_data

    def _get_data_for_combined_product(self, day_data, night_data):
        # Apply enhancements also to night-side data
        night_data = enhance2dataset(night_data)

        # Adjust bands so that they match
        # L/RGB -> RGB/RGB
        # LA/RGB -> RGBA/RGBA
        # RGB/RGBA -> RGBA/RGBA
        day_data = add_bands(day_data, night_data["bands"])
        night_data = add_bands(night_data, day_data["bands"])

        # Get merged metadata
        attrs = combine_metadata(day_data, night_data)

        return day_data, night_data, attrs

    def _mask_weights_with_data(
            self,
            weights: xr.DataArray,
            day_data: xr.DataArray,
            night_data: xr.DataArray,
    ) -> xr.DataArray:
        data_a = _get_single_channel(day_data)
        data_b = _get_single_channel(night_data)
        if "only" in self.day_night:
            mask = _get_weight_mask_for_single_side_product(data_a, data_b)
        else:
            mask = _get_weight_mask_for_daynight_product(weights, data_a, data_b)

        return weights.where(mask, np.nan)

    def _weight_data(
            self,
            day_data: xr.DataArray,
            night_data: xr.DataArray,
            weights: xr.DataArray,
            attrs: dict,
    ) -> list[xr.DataArray]:
        weights = self._fill_weights(weights)
        data = self._merge_bands_with_weights(day_data, night_data, weights, attrs)
        return data

    def _fill_weights(self, weights):
        if not self.include_alpha:
            fill = 1 if self.day_night == "night_only" else 0
            weights = weights.where(~np.isnan(weights), fill)
        return weights

    def _merge_bands_with_weights(self, day_data, night_data, weights, attrs):
        data = []
        for b in _get_band_names(day_data, night_data):
            day_band = _get_single_band_data(day_data, b)
            night_band = _get_single_band_data(night_data, b)
            # For day-only and night-only products only the alpha channel is weighted
            # If there's no alpha band, weight the actual data
            if self._is_weightable(b):
                day_band = day_band * weights
                night_band = night_band * (1 - weights)
            band = day_band + night_band
            band.attrs = attrs
            data.append(band)
        return data

    def _is_weightable(self, band):
        return (band == "A") or ("only" not in self.day_night) or not self.include_alpha


def _get_band_names(day_data, night_data):
    try:
        bands = day_data["bands"]
    except (IndexError, TypeError):
        bands = night_data["bands"]
    return bands


def _get_single_band_data(data, band):
    try:
        return data.sel(bands=band)
    except AttributeError:
        return data


def _get_single_channel(data: xr.DataArray) -> xr.DataArray:
    try:
        data = data[0, :, :]
        # remove coordinates that may be band-specific (ex. "bands")
        # and we don't care about anymore
        data = data.reset_coords(drop=True)
    except (IndexError, TypeError):
        pass
    return data


def _get_weight_mask_for_single_side_product(data_a, data_b):
    if data_b.shape:
        return ~da.isnan(data_b)
    return ~da.isnan(data_a)


def _get_weight_mask_for_daynight_product(weights, data_a, data_b):
    mask1 = (weights > 0) & ~np.isnan(data_a)
    mask2 = (weights < 1) & ~np.isnan(data_b)
    return mask1 | mask2


def add_alpha_bands(data):
    """Only used for DayNightCompositor.

    Add an alpha band to L or RGB composite as prerequisites for the following band matching
    to make the masked-out area transparent.
    """
    if "A" not in data["bands"].data:
        new_data = [data.sel(bands=band) for band in data["bands"].data]
        # Create alpha band based on a copy of the first "real" band
        alpha = new_data[0].copy()
        alpha.data = da.ones((data.sizes["y"],
                              data.sizes["x"]),
                             chunks=new_data[0].chunks,
                             dtype=data.dtype)
        # Rename band to indicate it's alpha
        alpha["bands"] = "A"
        new_data.append(alpha)
        new_data = xr.concat(new_data, dim="bands")
        new_data.attrs["mode"] = data.attrs["mode"] + "A"
        data = new_data
    return data


def zero_missing_data(data1, data2):
    """Replace NaN values with zeros in data1 if the data is valid in data2."""
    nans = np.logical_and(np.isnan(data1), np.logical_not(np.isnan(data2)))
    return data1.where(~nans, 0)


class BackgroundCompositor(GenericCompositor):
    """A compositor that overlays one composite on top of another.

    The output image mode will be determined by both foreground and background. Generally, when the background has
    an alpha band, the output image will also have one.

    ============  ============  ========
    Foreground     Background    Result
    ============  ============  ========
    L             L             L
    ------------  ------------  --------
    L             LA            LA
    ------------  ------------  --------
    L             RGB           RGB
    ------------  ------------  --------
    L             RGBA          RGBA
    ------------  ------------  --------
    LA            L             L
    ------------  ------------  --------
    LA            LA            LA
    ------------  ------------  --------
    LA            RGB           RGB
    ------------  ------------  --------
    LA            RGBA          RGBA
    ------------  ------------  --------
    RGB           L             RGB
    ------------  ------------  --------
    RGB           LA            RGBA
    ------------  ------------  --------
    RGB           RGB           RGB
    ------------  ------------  --------
    RGB           RGBA          RGBA
    ------------  ------------  --------
    RGBA          L             RGB
    ------------  ------------  --------
    RGBA          LA            RGBA
    ------------  ------------  --------
    RGBA          RGB           RGB
    ------------  ------------  --------
    RGBA          RGBA          RGBA
    ============  ============  ========

    """

    def __call__(self, projectables, *args, **kwargs):
        """Call the compositor."""
        projectables = self.match_data_arrays(projectables)

        # Get enhanced datasets
        foreground = enhance2dataset(projectables[0], convert_p=True)
        background = enhance2dataset(projectables[1], convert_p=True)
        before_bg_mode = background.attrs["mode"]

        # Adjust bands so that they have the same mode
        foreground = add_bands(foreground, background["bands"])
        background = add_bands(background, foreground["bands"])

        # It's important whether the alpha band of background is initially generated, e.g. by CloudCompositor
        # The result will be used to determine the output image mode
        initial_bg_alpha = "A" in before_bg_mode

        attrs = self._combine_metadata_with_mode_and_sensor(foreground, background)
        if "A" not in foreground.attrs["mode"] and "A" not in background.attrs["mode"]:
            data = self._simple_overlay(foreground, background)
        else:
            data = self._get_merged_image_data(foreground, background, initial_bg_alpha=initial_bg_alpha)
        for data_arr in data:
            data_arr.attrs = attrs
        res = super(BackgroundCompositor, self).__call__(data, **kwargs)
        attrs.update(res.attrs)
        res.attrs = attrs
        return res

    def _combine_metadata_with_mode_and_sensor(self,
                                               foreground: xr.DataArray,
                                               background: xr.DataArray
                                               ) -> dict:
        # Get merged metadata
        attrs = combine_metadata(foreground, background)
        # 'mode' is no longer valid after we've remove the 'A'
        # let the base class __call__ determine mode
        attrs.pop("mode", None)
        if attrs.get("sensor") is None:
            # sensor can be a set
            attrs["sensor"] = self._get_sensors([foreground, background])
        return attrs

    @staticmethod
    def _get_merged_image_data(foreground: xr.DataArray,
                               background: xr.DataArray,
                               initial_bg_alpha: bool,
                               ) -> list[xr.DataArray]:
        # For more info about alpha compositing please review https://en.wikipedia.org/wiki/Alpha_compositing
        alpha_fore = _get_alpha(foreground)
        alpha_back = _get_alpha(background)
        new_alpha = alpha_fore + alpha_back * (1 - alpha_fore)

        data = []

        # Pass the image data (alpha band will be dropped temporally) to the writer
        output_mode = background.attrs["mode"].replace("A", "")

        for band in output_mode:
            fg_band = foreground.sel(bands=band)
            bg_band = background.sel(bands=band)
            # Do the alpha compositing
            chan = (fg_band * alpha_fore + bg_band * alpha_back * (1 - alpha_fore)) / new_alpha
            # Fill the NaN area with background
            chan = xr.where(chan.isnull(), bg_band * alpha_back, chan)
            chan["bands"] = band
            data.append(chan)

        # If background has an initial alpha band, it will also be passed to the writer
        if initial_bg_alpha:
            new_alpha["bands"] = "A"
            data.append(new_alpha)

        return data

    @staticmethod
    def _simple_overlay(foreground: xr.DataArray,
                        background: xr.DataArray,) -> list[xr.DataArray]:
        # This is for the case when no alpha bands are involved
        # Just simply lay the foreground upon background
        data_arr = xr.where(foreground.isnull(), background, foreground)
        # Split to separate bands so the mode is correct
        data = [data_arr.sel(bands=b) for b in data_arr["bands"]]

        return data


def _get_alpha(dataset: xr.DataArray):
    # 1. This function is only used by _get_merged_image_data
    # 2. Both foreground and background have been through add_bands, so they have the same mode
    # 3. If none of them has alpha band, they will be passed to _simple_overlay not _get_merged_image_data
    # So any dataset(whether foreground or background) passed to this function has an alpha band for certain
    # We will use it directly
    alpha = dataset.sel(bands="A")
    # There could be NaNs in the alpha
    # Replace them with 0 to prevent cases like 1 + nan = nan, so they won't affect new_alpha
    alpha = xr.where(alpha.isnull(), 0, alpha)

    return alpha



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""ASCAT Soil moisture product reader for BUFR messages.

Based on the IASI L2 SO2 BUFR reader.

"""

import datetime as dt
import logging

import dask.array as da
import numpy as np
import xarray as xr

try:
    import eccodes as ec
except ImportError as e:
    raise ImportError(
        """Missing eccodes-python and/or eccodes C-library installation. Use conda to install eccodes.
           Error: """, e)

from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.utils import get_legacy_chunk_size

logger = logging.getLogger("AscatSoilMoistureBufr")

CHUNK_SIZE = get_legacy_chunk_size()


class AscatSoilMoistureBufr(BaseFileHandler):
    """File handler for the ASCAT Soil Moisture BUFR product."""

    def __init__(self, filename, filename_info, filetype_info, **kwargs):
        """Initialise the file handler for the ASCAT Soil Moisture BUFR data."""
        super(AscatSoilMoistureBufr, self).__init__(filename, filename_info, filetype_info)

        start_time, end_time = self.get_start_end_date()

        self.metadata = {}
        self.metadata["start_time"] = start_time
        self.metadata["end_time"] = end_time

    @property
    def start_time(self):
        """Return the start time of data acqusition."""
        return self.metadata["start_time"]

    @property
    def end_time(self):
        """Return the end time of data acquisition."""
        return self.metadata["end_time"]

    @property
    def platform_name(self):
        """Return spacecraft name."""
        return self.filename_info["platform"]

    def extract_msg_date_extremes(self, bufr, date_min=None, date_max=None):
        """Extract the minimum and maximum dates from a single bufr message."""
        ec.codes_set(bufr, "unpack", 1)
        size = ec.codes_get(bufr, "numberOfSubsets")
        years = np.resize(ec.codes_get_array(bufr, "year"), size)
        months = np.resize(ec.codes_get_array(bufr, "month"), size)
        days = np.resize(ec.codes_get_array(bufr, "day"), size)
        hours = np.resize(ec.codes_get_array(bufr, "hour"), size)
        minutes = np.resize(ec.codes_get_array(bufr, "minute"), size)
        seconds = np.resize(ec.codes_get_array(bufr, "second"), size)
        for year, month, day, hour, minute, second in zip(years, months, days, hours, minutes, seconds):
            time_stamp = dt.datetime(year, month, day, hour, minute, second)
            date_min = time_stamp if not date_min else min(date_min, time_stamp)
            date_max = time_stamp if not date_max else max(date_max, time_stamp)
        return date_min, date_max

    def get_start_end_date(self):
        """Get the first and last date from the bufr file."""
        with open(self.filename, "rb") as fh:
            date_min = None
            date_max = None
            while True:
                # get handle for message
                bufr = ec.codes_bufr_new_from_file(fh)
                if bufr is None:
                    break
                date_min, date_max = self.extract_msg_date_extremes(bufr, date_min, date_max)
            return date_min, date_max

    def get_bufr_data(self, key):
        """Get BUFR data by key."""
        attr = np.array([])
        with open(self.filename, "rb") as fh:
            while True:
                # get handle for message
                bufr = ec.codes_bufr_new_from_file(fh)
                if bufr is None:
                    break
                ec.codes_set(bufr, "unpack", 1)
                tmp = ec.codes_get_array(bufr, key, float)
                if len(tmp) == 1:
                    size = ec.codes_get(bufr, "numberOfSubsets")
                    tmp = np.resize(tmp, size)
                attr = np.append(attr, tmp)
                ec.codes_release(bufr)
        return attr

    def get_dataset(self, dataset_id, dataset_info):
        """Get dataset using the BUFR key in dataset_info."""
        arr = self.get_bufr_data(dataset_info["key"])
        if "fill_value" in dataset_info:
            arr[arr == dataset_info["fill_value"]] = np.nan
        arr = da.from_array(arr, chunks=CHUNK_SIZE)
        xarr = xr.DataArray(arr, dims=["y"], name=dataset_info["name"])
        xarr.attrs["platform_name"] = self.platform_name
        xarr.attrs.update(dataset_info)

        return xarr



# Copyright (c) 2017-2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""IASI-NG L2 reader implementation.

This reader supports reading all the products from the IASI-NG L2 processing
level:

* IASI-L2-TWV
* IASI-L2-CLD
* IASI-L2-GHG
* IASI-L2-SFC
* IASI-L2-O3
* IASI-L2-CO

For more information in the product files content, please refer to the
EPS-SG IASI-NG Level 2 Format specification document provided by EUMETSAT at:
https://user.eumetsat.int/s3/eup-strapi-media/EPS_SG_IASI_NG_L2_Product_Format_Specification_v4_B_cffb0a9273.pdf
"""

import re

import netCDF4
import numpy as np
import pandas as pd
import xarray as xr

from satpy.readers.core.netcdf import NetCDF4FsspecFileHandler


class IASINGL2NCFileHandler(NetCDF4FsspecFileHandler):
    """Reader for IASI-NG L2 products in NetCDF format."""

    def __init__(self, filename, filename_info, filetype_info, **kwargs):
        """Initialize object."""
        super().__init__(
            filename, filename_info, filetype_info, auto_maskandscale=True, **kwargs
        )

        self.sensors = {"iasi_ng"}

        self.dataset_infos = None
        self.variable_desc = {}
        self.dimensions_desc = {}

        patterns = self.filetype_info.get("ignored_patterns", [])
        self.ignored_patterns = [re.compile(pstr) for pstr in patterns]

        aliases = self.filetype_info.get("dataset_aliases", {})
        self.dataset_aliases = {re.compile(key): val for key, val in aliases.items()}

        self.register_available_datasets()

    @property
    def start_time(self):
        """Get the start time."""
        return self.filename_info["sensing_start_time"]

    @property
    def end_time(self):
        """Get the end time."""
        return self.filename_info["sensing_end_time"]

    @property
    def sensor_names(self):
        """List of sensors represented in this file."""
        return self.sensors

    # Note: patching the collect_groups_info method below to
    # also collect dimensions in sub groups.
    def _collect_groups_info(self, base_name, obj):
        for group_name, group_obj in obj.groups.items():
            full_group_name = base_name + group_name
            self.file_content[full_group_name] = group_obj
            self._collect_attrs(full_group_name, group_obj)
            self.collect_metadata(full_group_name, group_obj)
            self.collect_dimensions(full_group_name, group_obj)

    def available_datasets(self, configured_datasets=None):
        """Determine automatically the datasets provided by this file.

        First yield on any element from the provided configured_datasets,
        and then continues with the internally provided datasets.
        """
        for is_avail, ds_info in configured_datasets or []:
            yield is_avail, ds_info

        for _, ds_info in self.dataset_infos.items():
            yield True, ds_info

    def register_dataset(self, ds_name, desc):
        """Register a simple dataset given its name and a desc dict."""
        if ds_name in self.dataset_infos:
            raise KeyError(f"Dataset for {ds_name} already registered.")

        ds_infos = {
            "name": ds_name,
            "sensor": "iasi_ng",
            "file_type": self.filetype_info["file_type"],
        }

        ds_infos.update(desc)

        self.dataset_infos[ds_name] = ds_infos

    def same_dim_names_for_different_groups(self, dim_name, value):
        """Check if we already have this dim_name registered from another group."""
        return (
            dim_name in self.dimensions_desc and self.dimensions_desc[dim_name] != value
        )

    def process_dimension(self, key, value):
        """Process a dimension entry from the file_content."""
        dim_name = key.split("/")[-1]

        if self.same_dim_names_for_different_groups(dim_name, value):
            raise KeyError(f"Detected duplicated dim name: {dim_name}")

        self.dimensions_desc[dim_name] = value

    def has_variable_desc(self, var_path):
        """Check if a given variable path is available."""
        return var_path in self.variable_desc

    def process_attribute(self, key, value):
        """Process a attribute entry from the file_content."""
        var_path, aname = key.split("/attr/")

        if not self.has_variable_desc(var_path):
            return

        self.variable_desc[var_path]["attribs"][aname] = value

    def has_at_most_one_element(self, shape):
        """Check if a shape corresponds to an array with at most 1 element."""
        return np.prod(shape) <= 1

    def is_variable_ignored(self, var_name):
        """Check if a variable should be ignored."""
        return any(p.search(var_name) is not None for p in self.ignored_patterns)

    def prepare_variable_description(self, key, shape):
        """Prepare a description for a given variable."""
        prefix, var_name = key.rsplit("/", 1)
        dims = self.file_content[f"{key}/dimensions"]
        dtype = self.file_content[f"{key}/dtype"]

        return {
            "location": key,
            "prefix": prefix,
            "var_name": var_name,
            "shape": shape,
            "dtype": f"{dtype}",
            "dims": dims,
            "attribs": {},
        }

    def process_variable(self, key):
        """Process a variable entry from the file_content."""
        shape = self.file_content[f"{key}/shape"]

        if self.has_at_most_one_element(shape):
            return

        if self.is_variable_ignored(key):
            return

        self.variable_desc[key] = self.prepare_variable_description(key, shape)

    def parse_file_content(self):
        """Parse the file_content to discover the available datasets and dimensions."""
        for key, val in self.file_content.items():

            if "/dimension/" in key:
                self.process_dimension(key, val)
                continue

            if "/attr/" in key:
                self.process_attribute(key, val)
                continue

            if f"{key}/shape" in self.file_content:
                self.process_variable(key)
                continue

    def check_variable_alias(self, vpath, ds_name):
        """Check if a variable path matches an alias pattern."""
        for pat, sub in self.dataset_aliases.items():
            match = pat.search(vpath)
            if match:
                var_name = match.group(1)
                return sub.replace("${VAR_NAME}", var_name)

        return ds_name

    def register_available_datasets(self):
        """Register the available dataset in the current product file."""
        if self.dataset_infos is not None:
            return

        self.dataset_infos = {}

        self.parse_file_content()

        for vpath, desc in self.variable_desc.items():
            ds_name = desc["var_name"]
            ds_name = self.check_variable_alias(vpath, ds_name)

            unit = desc["attribs"].get("units", None)
            if unit is not None and unit.startswith("seconds since "):
                desc["seconds_since_epoch"] = unit.replace("seconds since ", "")

            self.register_dataset(ds_name, desc)

    def get_dataset_infos(self, ds_name):
        """Retrieve the dataset infos corresponding to one of the registered datasets."""
        if ds_name not in self.dataset_infos:
            raise KeyError(f"No dataset registered for {ds_name}")

        return self.dataset_infos[ds_name]

    def is_attribute_path(self, var_path):
        """Check if a given path is a root attribute path."""
        return var_path.startswith("/attr")

    def is_property_path(self, var_path):
        """Check if a given path is a sub-property path."""
        return var_path.endswith(("/dtype", "/shape", "/dimensions"))

    def is_netcdf_group(self, obj):
        """Check if a given object is a netCDF group."""
        return isinstance(obj, netCDF4.Group)

    def variable_path_exists(self, var_path):
        """Check if a given variable path is available in the underlying netCDF file.

        All we really need to do here is to access the file_content dictionary
        and check if we have a variable under that var_path key.
        """
        if self.is_attribute_path(var_path) or self.is_property_path(var_path):
            return False

        if var_path in self.file_content:
            return not self.is_netcdf_group(self.file_content[var_path])

        return False

    def convert_to_datetime(self, data_array, ds_info):
        """Convert the data to datetime values."""
        epoch = ds_info["seconds_since_epoch"]

        # Note: converting the time values to ns precision to avoid warnings
        # from panda+numpy:
        base_time = np.datetime64(pd.to_datetime(epoch), "ns")
        nanoseconds = data_array.astype("timedelta64[ns]") * 1e9

        data_array = xr.DataArray(
            data=base_time + nanoseconds,
            dims=data_array.dims,
            attrs=data_array.attrs,
        )

        return data_array

    def get_transformed_dataset(self, ds_info):
        """Retrieve a dataset with all transformations applied on it."""
        vname = ds_info["location"]

        if not self.variable_path_exists(vname):
            raise KeyError(f"Invalid variable path: {vname}")

        arr = self[vname]

        if "seconds_since_epoch" in ds_info:
            arr = self.convert_to_datetime(arr, ds_info)

        return arr

    def get_dataset(self, dataset_id, ds_info=None):
        """Get a dataset."""
        ds_name = dataset_id["name"]

        if ds_name not in self.dataset_infos:
            return None

        if ds_info is None:
            ds_info = self.get_dataset_infos(ds_name)

        ds_name = ds_info["name"]

        data_array = self.get_transformed_dataset(ds_info)

        return data_array



"""Implementation of some image filters."""

import logging

import dask.array as da
import numpy as np
import xarray as xr

from satpy.modifiers.filters import Median


def test_median(caplog):
    """Test the median filter modifier."""
    caplog.set_level(logging.DEBUG)
    dims = "y", "x"
    coordinates = dict(x=np.arange(6), y=np.arange(6))
    attrs = dict(units="K")
    median_filter_params = dict(size=3)
    name = "median_filter"
    median_filter = Median(median_filter_params, name=name)
    array = xr.DataArray(da.arange(36).reshape((6, 6)), coords=coordinates, dims=dims, attrs=attrs)
    res = median_filter([array])
    filtered_array = np.array([[1, 2, 3, 4, 5, 5],
                               [6, 7, 8, 9, 10, 11],
                               [12, 13, 14, 15, 16, 17],
                               [18, 19, 20, 21, 22, 23],
                               [24, 25, 26, 27, 28, 29],
                               [30, 30, 31, 32, 33, 34]])
    np.testing.assert_allclose(res, filtered_array)
    assert res.dims == dims
    assert attrs.items() <= res.attrs.items()
    assert res.attrs["name"] == name
    np.testing.assert_equal(res.coords["x"], coordinates["x"])
    np.testing.assert_equal(res.coords["y"], coordinates["y"])
    assert "Apply median filtering with parameters {'size': 3}" in caplog.text



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the goes imager nc reader (NOAA CLASS variant)."""

import datetime
import unittest
from unittest import mock

import numpy as np
import pytest
import xarray as xr
from pyresample.geometry import AreaDefinition

from satpy.readers.goes_imager_nc import is_vis_channel
from satpy.tests.utils import make_dataid

# NOTE:
# The following fixtures are not defined in this file, but are used and injected by Pytest:
# - request


class GOESNCBaseFileHandlerTest(unittest.TestCase):
    """Testing the file handler."""

    longMessage = True

    @mock.patch("satpy.readers.goes_imager_nc.xr")
    @mock.patch.multiple("satpy.readers.goes_imager_nc.GOESNCBaseFileHandler",
                         _get_sector=mock.MagicMock())
    def setUp(self, xr_):
        """Set up the tests."""
        from satpy.readers.goes_imager_nc import CALIB_COEFS, GOESNCBaseFileHandler

        self.coefs = CALIB_COEFS["GOES-15"]

        # Mock file access to return a fake dataset.
        self.time = datetime.datetime(2018, 8, 16, 16, 7)
        self.dummy3d = np.zeros((1, 2, 2))
        self.dummy2d = np.zeros((2, 2))
        self.band = 1
        self.nc = xr.Dataset(
            {"data": xr.DataArray(self.dummy3d, dims=("time", "yc", "xc")),
             "lon": xr.DataArray(data=self.dummy2d, dims=("yc", "xc")),
             "lat": xr.DataArray(data=self.dummy2d, dims=("yc", "xc")),
             "time": xr.DataArray(data=np.array([self.time],
                                                dtype="datetime64[ns]"),
                                  dims=("time",)),
             "bands": xr.DataArray(data=np.array([self.band]))},
            attrs={"Satellite Sensor": "G-15"})
        xr_.open_dataset.return_value = self.nc

        # Instantiate reader using the mocked open_dataset() method. Also, make
        # the reader believe all abstract methods have been implemented.
        self.reader = GOESNCBaseFileHandler(filename="dummy", filename_info={},
                                            filetype_info={})

    def test_init(self):
        """Tests reader initialization."""
        assert self.reader.nlines == self.dummy2d.shape[0]
        assert self.reader.ncols == self.dummy2d.shape[1]
        assert self.reader.platform_name == "GOES-15"
        assert self.reader.platform_shortname == "goes15"
        assert self.reader.gvar_channel == self.band
        assert isinstance(self.reader.geo_data, xr.Dataset)

    def test_get_nadir_pixel(self):
        """Test identification of the nadir pixel."""
        from satpy.readers.goes_imager_nc import FULL_DISC

        earth_mask = np.array([[0, 0, 0, 0],
                               [0, 1, 0, 0],
                               [1, 1, 1, 0],
                               [0, 1, 0, 0],
                               [0, 0, 0, 0]])
        nadir_row, nadir_col = self.reader._get_nadir_pixel(
            earth_mask=earth_mask, sector=FULL_DISC)
        assert (nadir_row, nadir_col) == (2, 1), "Incorrect nadir pixel"

    def test_viscounts2radiance(self):
        """Test conversion from VIS counts to radiance."""
        # Reference data is for detector #1
        slope = self.coefs["00_7"]["slope"][0]
        offset = self.coefs["00_7"]["offset"][0]
        counts = xr.DataArray([0, 100, 200, 500, 1000, 1023])
        rad_expected = xr.DataArray(
            [0., 41.54896, 100.06862,
             275.6276, 568.2259, 581.685422])
        rad = self.reader._viscounts2radiance(counts=counts, slope=slope,
                                              offset=offset)
        assert np.allclose(rad.data, rad_expected.data, atol=1e-06), "Incorrect conversion from VIS counts to radiance"

    def test_ircounts2radiance(self):
        """Test conversion from IR counts to radiance."""
        # Test counts
        counts = xr.DataArray([0, 100, 500, 1000, 1023])

        # Reference Radiance from NOAA lookup tables (same for detectors 1 and
        # 2, see [IR])
        rad_expected = {
            "03_9": np.array([0, 0.140, 1.899, 4.098, 4.199]),
            "06_5": np.array([0, 1.825, 12.124, 24.998, 25.590]),
            "10_7": np.array([0, 16.126, 92.630, 188.259, 192.658]),
            "13_3": np.array([0, 15.084, 87.421, 177.842, 182.001])
        }

        # The input counts are exact, but the accuracy of the output radiance is
        # limited to 3 digits
        atol = 1E-3

        for ch in sorted(rad_expected.keys()):
            coefs = self.coefs[ch]
            rad = self.reader._ircounts2radiance(
                counts=counts, scale=coefs["scale"], offset=coefs["offset"])
            assert np.allclose(rad.data, rad_expected[ch], atol=atol), \
                "Incorrect conversion from IR counts to radiance in channel {}".format(ch)

    def test_calibrate_vis(self):
        """Test VIS calibration."""
        rad = xr.DataArray([0, 1, 10, 100, 500])
        refl_expected = xr.DataArray([0., 0.188852, 1.88852, 18.8852, 94.426])
        refl = self.reader._calibrate_vis(radiance=rad,
                                          k=self.coefs["00_7"]["k"])
        assert np.allclose(refl.data, refl_expected.data, atol=1e-06), \
            "Incorrect conversion from radiance to reflectance"

    def test_calibrate_ir(self):
        """Test IR calibration."""
        # Test radiance values and corresponding BT from NOAA lookup tables
        # rev. H (see [IR]).
        rad = {
            "03_9": xr.DataArray([0, 0.1, 2, 3.997, 4.199]),
            "06_5": xr.DataArray([0, 0.821, 12.201, 25.590, 100]),
            "10_7": xr.DataArray([0, 11.727, 101.810, 189.407, 192.658]),
            "13_3": xr.DataArray([0, 22.679, 90.133, 182.001, 500])
        }
        bt_expected = {
            "03_9": np.array([[np.nan, 253.213, 319.451, 339.983, np.nan],
                              [np.nan, 253.213, 319.451, 339.983, np.nan]]),
            "06_5": np.array([[np.nan, 200.291, 267.860, 294.988, np.nan],
                              [np.nan, 200.308, 267.879, 295.008, np.nan]]),
            "10_7": np.array([[np.nan, 200.105, 294.437, 339.960, np.nan],
                              [np.nan, 200.097, 294.429, 339.953, np.nan]]),
            "13_3": np.array([[np.nan, 200.006, 267.517, 321.986, np.nan],
                              [np.nan, 200.014, 267.524, 321.990, np.nan]])
        }  # first row is for detector 1, second for detector 2.

        # The accuracy of the input radiance is limited to 3 digits so that
        # the results differ slightly.
        atol = {"03_9": 0.04, "06_5": 0.03, "10_7": 0.01, "13_3": 0.01}

        for ch in sorted(rad.keys()):
            coefs = self.coefs[ch]
            for det in [0, 1]:
                bt = self.reader._calibrate_ir(radiance=rad[ch],
                                               coefs={"a": coefs["a"][det],
                                                      "b": coefs["b"][det],
                                                      "n": coefs["n"][det],
                                                      "btmin": coefs["btmin"],
                                                      "btmax": coefs["btmax"]})
                assert np.allclose(bt.data, bt_expected[ch][det], equal_nan=True, atol=atol[ch]), \
                    f"Incorrect conversion from radiance to brightness temperature in channel {ch} detector {det}"

    def test_start_time(self):
        """Test dataset start time stamp."""
        assert self.reader.start_time == self.time

    def test_end_time(self):
        """Test dataset end time stamp."""
        from satpy.readers.goes_imager_nc import FULL_DISC, SCAN_DURATION, UNKNOWN_SECTOR
        expected = {
            UNKNOWN_SECTOR: self.time,
            FULL_DISC: self.time + SCAN_DURATION[FULL_DISC]
        }
        for sector, end_time in expected.items():
            self.reader.sector = sector
            assert self.reader.end_time == end_time


class TestMetadata:
    """Testcase for dataset metadata."""

    @pytest.fixture(params=[1, 2])
    def channel_id(self, request):
        """Set channel ID."""
        return request.param

    @pytest.fixture(params=[True, False])
    def yaw_flip(self, request):
        """Set yaw-flip flag."""
        return request.param

    def _apply_yaw_flip(self, data_array, yaw_flip):
        if yaw_flip:
            data_array.data = np.flipud(data_array.data)
        return data_array

    @pytest.fixture
    def lons_lats(self, yaw_flip):
        """Get longitudes and latitudes."""
        lon = xr.DataArray(
            [[-1, 0, 1, 2],
             [-1, 0, 1, 2],
             [-1, 0, 1, 2]],
            dims=("yc", "xc")
        )
        lat = xr.DataArray(
            [[9999, 9999, 9999, 9999],
             [1, 1, 1, 1],
             [-1, -1, -1, -1]],
            dims=("yc", "xc")
        )
        self._apply_yaw_flip(lat, yaw_flip)
        return lon, lat

    @pytest.fixture
    def dataset(self, lons_lats, channel_id):
        """Create a fake dataset."""
        lon, lat = lons_lats
        data = xr.DataArray(
            [[[1, 2, 3, 4],
              [5, 6, 7, 8],
              [9, 10, 11, 12]]],
            dims=("time", "yc", "xc")
        )
        time = xr.DataArray(
            [np.datetime64("2018-01-01 12:00:00").astype("datetime64[ns]")],
            dims="time"
        )
        bands = xr.DataArray([channel_id], dims="bands")
        return xr.Dataset(
            {
                "data": data,
                "lon": lon,
                "lat": lat,
                "time": time,
                "bands": bands,
            },
            attrs={"Satellite Sensor": "G-15"}
        )

    @pytest.fixture
    def earth_mask(self, yaw_flip):
        """Get expected earth mask."""
        earth_mask = xr.DataArray(
            [[False, False, False, False],
             [True, True, True, True],
             [True, True, True, True]],
            dims=("yc", "xc"),
        )
        self._apply_yaw_flip(earth_mask, yaw_flip)
        return earth_mask

    @pytest.fixture
    def geometry(self, channel_id, yaw_flip):
        """Get expected geometry."""
        shapes = {
            1: {"width": 10847, "height": 10810},
            2: {"width": 2712, "height": 2702}
        }
        return {
            "nadir_row": 0 if yaw_flip else 1,
            "projection_longitude": -1 if yaw_flip else 1,
            "shape": shapes[channel_id]
        }

    @pytest.fixture
    def expected(self, geometry, earth_mask, yaw_flip):
        """Define expected metadata."""
        proj_dict = {
            "a": "6378169",
            "h": "35785831",
            "lon_0": "0",
            "no_defs": "None",
            "proj": "geos",
            "rf": "295.488065897001",
            "type": "crs",
            "units": "m",
            "x_0": "0",
            "y_0": "0"
        }
        area = AreaDefinition(
            area_id="goes_geos_uniform",
            proj_id="goes_geos_uniform",
            description="GOES-15 geostationary projection (uniform sampling)",
            projection=proj_dict,
            area_extent=(-5434201.1352, -5415668.5992, 5434201.1352, 5415668.5992),
            **geometry["shape"]
        )
        return {
            "area_def_uni": area,
            "earth_mask": earth_mask,
            "yaw_flip": yaw_flip,
            "lon0": 0,
            "lat0": geometry["projection_longitude"],
            "nadir_row": geometry["nadir_row"],
            "nadir_col": 1
        }

    @pytest.fixture
    def mocked_file_handler(self, dataset):
        """Mock file handler to load the given fake dataset."""
        from satpy.readers.goes_imager_nc import FULL_DISC, GOESNCFileHandler
        with mock.patch("satpy.readers.goes_imager_nc.xr") as xr_:
            xr_.open_dataset.return_value = dataset
            GOESNCFileHandler.vis_sectors[(3, 4)] = FULL_DISC
            GOESNCFileHandler.ir_sectors[(3, 4)] = FULL_DISC
            GOESNCFileHandler.yaw_flip_sampling_distance = 1
            return GOESNCFileHandler(
                filename="dummy",
                filename_info={},
                filetype_info={},
            )

    def test_metadata(self, mocked_file_handler, expected):
        """Test dataset metadata."""
        metadata = mocked_file_handler.meta
        self._assert_earth_mask_equal(metadata, expected)
        assert metadata == expected

    def _assert_earth_mask_equal(self, metadata, expected):
        earth_mask_tst = metadata.pop("earth_mask")
        earth_mask_ref = expected.pop("earth_mask")
        xr.testing.assert_allclose(earth_mask_tst, earth_mask_ref)


class GOESNCFileHandlerTest(unittest.TestCase):
    """Test the file handler."""

    longMessage = True

    @mock.patch("satpy.readers.goes_imager_nc.xr")
    def setUp(self, xr_):
        """Set up the tests."""
        from satpy.readers.goes_imager_nc import CALIB_COEFS, GOESNCFileHandler

        self.coefs = CALIB_COEFS["GOES-15"]
        self.all_coefs = CALIB_COEFS
        self.channels = sorted(self.coefs.keys())
        self.ir_channels = sorted([ch for ch in self.channels
                                   if not is_vis_channel(ch)])
        self.vis_channels = sorted([ch for ch in self.channels
                                    if is_vis_channel(ch)])

        # Mock file access to return a fake dataset. Choose a medium count value
        # (100) to avoid elements being masked due to invalid
        # radiance/reflectance/BT
        nrows = ncols = 300
        self.counts = 100 * 32 * np.ones((1, nrows, ncols))  # emulate 10-bit
        self.lon = np.zeros((nrows, ncols))  # Dummy
        self.lat = np.repeat(np.linspace(-150, 150, nrows), ncols).reshape(
            nrows, ncols)  # Includes invalid values to be masked

        xr_.open_dataset.return_value = xr.Dataset(
            {"data": xr.DataArray(data=self.counts, dims=("time", "yc", "xc")),
             "lon": xr.DataArray(data=self.lon, dims=("yc", "xc")),
             "lat": xr.DataArray(data=self.lat, dims=("yc", "xc")),
             "time": xr.DataArray(data=np.array([0], dtype="datetime64[ns]"),
                                  dims=("time",)),
             "bands": xr.DataArray(data=np.array([1]))},
            attrs={"Satellite Sensor": "G-15"})

        # Instantiate reader using the mocked open_dataset() method
        self.reader = GOESNCFileHandler(filename="dummy", filename_info={},
                                        filetype_info={})

    def test_get_dataset_coords(self):
        """Test whether coordinates returned by get_dataset() are correct."""
        lon = self.reader.get_dataset(key=make_dataid(name="longitude"),
                                      info={})
        lat = self.reader.get_dataset(key=make_dataid(name="latitude"),
                                      info={})
        # ... this only compares the valid (unmasked) elements
        assert np.all(lat.to_masked_array() == self.lat), "get_dataset() returns invalid latitude"
        assert np.all(lon.to_masked_array() == self.lon), "get_dataset() returns invalid longitude"

    def test_get_dataset_counts(self):
        """Test whether counts returned by get_dataset() are correct."""
        from satpy.readers.goes_imager_nc import ALTITUDE, UNKNOWN_SECTOR

        self.reader.meta.update({"lon0": -75.0,
                                 "lat0": 0.0,
                                 "sector": UNKNOWN_SECTOR,
                                 "nadir_row": 1,
                                 "nadir_col": 2,
                                 "area_def_uni": "some_area"})
        attrs_exp = {"orbital_parameters": {"projection_longitude": -75.0,
                                            "projection_latitude": 0.0,
                                            "projection_altitude": ALTITUDE,
                                            "yaw_flip": True},
                     "platform_name": "GOES-15",
                     "sensor": "goes_imager",
                     "sector": UNKNOWN_SECTOR,
                     "nadir_row": 1,
                     "nadir_col": 2,
                     "area_def_uniform_sampling": "some_area"}

        for ch in self.channels:
            counts = self.reader.get_dataset(
                key=make_dataid(name=ch, calibration="counts"), info={})
            # ... this only compares the valid (unmasked) elements
            assert np.all(self.counts / 32.0 == counts.to_masked_array()), \
                f"get_dataset() returns invalid counts for channel {ch}"

            # Check attributes
            assert counts.attrs == attrs_exp

    def test_get_dataset_masks(self):
        """Test whether data and coordinates are masked consistently."""
        # Requires that no element has been masked due to invalid
        # radiance/reflectance/BT (see setUp()).
        lon = self.reader.get_dataset(key=make_dataid(name="longitude"),
                                      info={})
        lon_mask = lon.to_masked_array().mask
        for ch in self.channels:
            for calib in ("counts", "radiance", "reflectance",
                          "brightness_temperature"):
                try:
                    data = self.reader.get_dataset(
                        key=make_dataid(name=ch, calibration=calib), info={})
                except ValueError:
                    continue
                data_mask = data.to_masked_array().mask
                assert np.all(data_mask == lon_mask), \
                    f"get_dataset() returns inconsistently masked {calib} in channel {ch}"

    def test_get_dataset_invalid(self):
        """Test handling of invalid calibrations."""
        # VIS -> BT
        args = dict(key=make_dataid(name="00_7",
                                    calibration="brightness_temperature"),
                    info={})
        with pytest.raises(ValueError, match="Cannot calibrate VIS channel to 2"):
            self.reader.get_dataset(**args)

        # IR -> Reflectance
        args = dict(key=make_dataid(name="10_7",
                                    calibration="reflectance"),
                    info={})
        with pytest.raises(ValueError, match="Cannot calibrate IR channel to 1"):
            self.reader.get_dataset(**args)

        # Unsupported calibration
        with pytest.raises(ValueError, match="invalid invalid value for <enum 'calibration'>"):
            _ = dict(key=make_dataid(name="10_7",
                                     calibration="invalid"),
                     info={})

    def test_calibrate(self):
        """Test whether the correct calibration methods are called."""
        for ch in self.channels:
            if is_vis_channel(ch):
                calibs = {"radiance": "_viscounts2radiance",
                          "reflectance": "_calibrate_vis"}
            else:
                calibs = {"radiance": "_ircounts2radiance",
                          "brightness_temperature": "_calibrate_ir"}
            for calib, method in calibs.items():
                with mock.patch.object(self.reader, method) as target_func:
                    self.reader.calibrate(counts=self.reader.nc["data"],
                                          calibration=calib, channel=ch)
                    target_func.assert_called()

    def test_get_sector(self):
        """Test sector identification."""
        from satpy.readers.goes_imager_nc import (
            FULL_DISC,
            NORTH_HEMIS_EAST,
            NORTH_HEMIS_WEST,
            SOUTH_HEMIS_EAST,
            SOUTH_HEMIS_WEST,
            UNKNOWN_SECTOR,
        )
        shapes_vis = {
            (10800, 20754): FULL_DISC,
            (7286, 13900): NORTH_HEMIS_EAST,
            (2301, 13840): SOUTH_HEMIS_EAST,
            (5400, 13200): NORTH_HEMIS_WEST,
            (4300, 11090): SOUTH_HEMIS_WEST,
            (123, 456): UNKNOWN_SECTOR
        }
        shapes_ir = {
            (2700, 5200): FULL_DISC,
            (1850, 3450): NORTH_HEMIS_EAST,
            (600, 3500): SOUTH_HEMIS_EAST,
            (1310, 3300): NORTH_HEMIS_WEST,
            (1099, 2800): SOUTH_HEMIS_WEST,
            (123, 456): UNKNOWN_SECTOR
        }
        shapes = shapes_ir.copy()
        shapes.update(shapes_vis)
        for (nlines, ncols), sector_ref in shapes.items():
            if (nlines, ncols) in shapes_vis:
                channel = "00_7"
            else:
                channel = "10_7"
            sector = self.reader._get_sector(channel=channel, nlines=nlines,
                                             ncols=ncols)
            assert sector == sector_ref, "Incorrect sector identification"


class TestChannelIdentification:
    """Test identification of channel type."""

    @pytest.mark.parametrize(
        ("channel_name", "expected"),
        [
            ("00_7", True),
            ("10_7", False),
            (1, True),
            (2, False)
        ]
    )
    def test_is_vis_channel(self, channel_name, expected):
        """Test vis channel identification."""
        assert is_vis_channel(channel_name) == expected

    def test_invalid_channel(self):
        """Test handling of invalid channel type."""
        with pytest.raises(ValueError, match="Invalid channel"):
            is_vis_channel({"foo": "bar"})



# Copyright (c) 2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Test enhancer class."""
from __future__ import annotations

import os
import pathlib

import numpy as np
import pytest
import xarray as xr
from dask import array as da
from trollimage.xrimage import XRImage

from satpy.enhancements.enhancer import Enhancer
from satpy.writers.core.image import ImageWriter


class TestEnhancer:
    """Test basic `Enhancer` functionality with builtin configs."""

    def test_basic_init_no_args(self):
        """Test Enhancer init with no arguments passed."""
        e = Enhancer()
        assert e.enhancement_tree is not None

    def test_basic_init_no_enh(self):
        """Test Enhancer init requesting no enhancements."""
        e = Enhancer(enhancement_config_file=False)
        assert e.enhancement_tree is None

    def test_basic_init_provided_enh(self):
        """Test Enhancer init with string enhancement configs."""
        e = Enhancer(enhancement_config_file=["""enhancements:
  enh1:
    standard_name: toa_bidirectional_reflectance
    operations:
    - name: stretch
      method: !!python/name:satpy.enhancements.contrast.stretch
      kwargs: {stretch: linear}
"""])
        assert e.enhancement_tree is not None

    def test_init_nonexistent_enh_file(self):
        """Test Enhancer init with a nonexistent enhancement configuration file."""
        with pytest.raises(ValueError, match="YAML file doesn't exist or string is not YAML dict:.*"):
            Enhancer(enhancement_config_file="is_not_a_valid_filename_?.yaml")

    def test_print_tree(self, capsys):
        """Test enhancement decision tree printing."""
        enh = Enhancer()
        enh.enhancement_tree.print_tree()
        stdout = capsys.readouterr().out
        lines = stdout.splitlines()
        assert lines[0].startswith("name=<wildcard>")
        # make sure lines are indented
        assert lines[1].startswith("  reader=")
        assert lines[2].startswith("    platform_name=")


def test_xrimage_1d():
    """Conversion to image."""
    p = xr.DataArray(np.arange(25), dims=["y"])
    with pytest.raises(ValueError, match="Data must have a 'y' and 'x' dimension"):
        XRImage(p)


def test_xrimage_2d():
    """Conversion to image."""
    data = np.arange(25).reshape((5, 5))
    p = xr.DataArray(data, attrs=dict(mode="L", fill_value=0,
                                      palette=[0, 1, 2, 3, 4, 5]),
                     dims=["y", "x"])
    XRImage(p)


def test_xrimage_3d():
    """Conversion to image."""
    data = np.arange(75).reshape((3, 5, 5))
    p = xr.DataArray(data, dims=["bands", "y", "x"])
    p["bands"] = ["R", "G", "B"]
    XRImage(p)


class _CustomImageWriter(ImageWriter):
    def __init__(self, **kwargs):
        super().__init__(name="test", config_files=[], **kwargs)
        self.img = None

    def save_image(self, img, **kwargs):
        self.img = img
        return [None]


class _BaseCustomEnhancementConfigTests:

    TEST_CONFIGS: dict[str, str] = {}

    @pytest.fixture(scope="class", autouse=True)
    def test_configs_path(self, tmp_path_factory):
        """Create test enhancement configuration files in a temporary directory.

        The root temporary directory is changed to and returned.

        """
        prev_cwd = pathlib.Path.cwd()
        tmp_path = tmp_path_factory.mktemp("config")
        os.chdir(tmp_path)

        for fn, content in self.TEST_CONFIGS.items():
            config_rel_dir = os.path.dirname(fn)
            if config_rel_dir:
                os.makedirs(config_rel_dir, exist_ok=True)
            with open(fn, "w") as f:
                f.write(content)

        try:
            yield tmp_path
        finally:
            os.chdir(prev_cwd)


class TestComplexSensorEnhancerConfigs(_BaseCustomEnhancementConfigTests):
    """Test enhancement configs that use or expect multiple sensors."""

    ENH_FN = "test_sensor1.yaml"
    ENH_FN2 = "test_sensor2.yaml"

    TEST_CONFIGS = {
        ENH_FN: """
enhancements:
  test1_sensor1_specific:
    name: test1
    sensor: test_sensor1
    operations:
    - name: stretch
      method: !!python/name:satpy.enhancements.contrast.stretch
      kwargs: {stretch: crude, min_stretch: 0, max_stretch: 200}

        """,
        ENH_FN2: """
enhancements:
  default:
    operations:
    - name: stretch
      method: !!python/name:satpy.enhancements.contrast.stretch
      kwargs: {stretch: crude, min_stretch: 0, max_stretch: 100}
  test1_sensor2_specific:
    name: test1
    sensor: test_sensor2
    operations:
    - name: stretch
      method: !!python/name:satpy.enhancements.contrast.stretch
      kwargs: {stretch: crude, min_stretch: 0, max_stretch: 50}
  exact_multisensor_comp:
    name: my_comp
    sensor: [test_sensor1, test_sensor2]
    operations:
    - name: stretch
      method: !!python/name:satpy.enhancements.contrast.stretch
      kwargs: {stretch: crude, min_stretch: 0, max_stretch: 20}
            """,
    }

    def test_multisensor_choice(self, test_configs_path):
        """Test that a DataArray with two sensors works."""
        from xarray import DataArray

        from satpy.enhancements.enhancer import Enhancer, get_enhanced_image
        ds = DataArray(np.arange(1, 11.).reshape((2, 5)),
                       attrs={
                           "name": "test1",
                           "sensor": {"test_sensor2", "test_sensor1"},
                           "mode": "L"
                       },
                       dims=["y", "x"])
        e = Enhancer()
        assert e.enhancement_tree is not None
        img = get_enhanced_image(ds, enhance=e)
        # make sure that both sensor configs were loaded
        assert (set(pathlib.Path(config) for config in e.sensor_enhancement_configs) ==
                {test_configs_path / self.ENH_FN,
                 test_configs_path / self.ENH_FN2})
        # test_sensor1 config should have been used because it is
        # alphabetically first
        np.testing.assert_allclose(img.data.values[0], ds.data / 200.0)

    def test_multisensor_exact(self, test_configs_path):
        """Test that a DataArray with two sensors can match exactly."""
        from xarray import DataArray

        from satpy.enhancements.enhancer import Enhancer, get_enhanced_image
        ds = DataArray(np.arange(1, 11.).reshape((2, 5)),
                       attrs={
                           "name": "my_comp",
                           "sensor": {"test_sensor2", "test_sensor1"},
                           "mode": "L"
                       },
                       dims=["y", "x"])
        e = Enhancer()
        assert e.enhancement_tree is not None
        img = get_enhanced_image(ds, enhance=e)
        # make sure that both sensor configs were loaded
        assert (set(pathlib.Path(config) for config in e.sensor_enhancement_configs) ==
                {test_configs_path / self.ENH_FN,
                 test_configs_path / self.ENH_FN2})
        # test_sensor1 config should have been used because it is
        # alphabetically first
        np.testing.assert_allclose(img.data.values[0], ds.data / 20.0)

    def test_enhance_bad_query_value(self):
        """Test Enhancer doesn't fail when query includes bad values."""
        from xarray import DataArray

        from satpy.enhancements.enhancer import Enhancer, get_enhanced_image
        ds = DataArray(np.arange(1, 11.).reshape((2, 5)),
                       attrs=dict(name=["I", "am", "invalid"], sensor="test_sensor2", mode="L"),
                       dims=["y", "x"])
        e = Enhancer()
        assert e.enhancement_tree is not None
        with pytest.raises(KeyError, match="No .* found for None"):
            get_enhanced_image(ds, enhance=e)


class TestEnhancerUserConfigs(_BaseCustomEnhancementConfigTests):
    """Test `Enhancer` functionality when user's custom configurations are present."""

    ENH_FN = "test_sensor.yaml"
    ENH_ENH_FN = os.path.join("enhancements", ENH_FN)
    ENH_FN2 = "test_sensor2.yaml"
    ENH_ENH_FN2 = os.path.join("enhancements", ENH_FN2)
    ENH_FN3 = "test_empty.yaml"

    TEST_CONFIGS = {
        ENH_FN: """
enhancements:
  test1_default:
    name: test1
    operations:
    - name: stretch
      method: !!python/name:satpy.enhancements.contrast.stretch
      kwargs: {stretch: linear, cutoffs: [0., 0.]}

        """,
        ENH_ENH_FN: """
enhancements:
  test1_kelvin:
    name: test1
    units: kelvin
    operations:
    - name: stretch
      method: !!python/name:satpy.enhancements.contrast.stretch
      kwargs: {stretch: crude, min_stretch: 0, max_stretch: 20}

        """,
        ENH_FN2: """


        """,
        ENH_ENH_FN2: """

        """,
        ENH_FN3: """""",
    }

    def test_enhance_empty_config(self, test_configs_path):
        """Test Enhancer doesn't fail with empty enhancement file."""
        from xarray import DataArray

        from satpy.enhancements.enhancer import Enhancer, get_enhanced_image
        ds = DataArray(np.arange(1, 11.).reshape((2, 5)),
                       attrs=dict(sensor="test_empty", mode="L"),
                       dims=["y", "x"])
        e = Enhancer()
        assert e.enhancement_tree is not None
        get_enhanced_image(ds, enhance=e)
        assert (set(pathlib.Path(config) for config in e.sensor_enhancement_configs) ==
                {test_configs_path / self.ENH_FN3})

    def test_enhance_with_sensor_no_entry(self, test_configs_path):
        """Test enhancing an image that has no configuration sections."""
        from xarray import DataArray

        from satpy.enhancements.enhancer import Enhancer, get_enhanced_image
        ds = DataArray(np.arange(1, 11.).reshape((2, 5)),
                       attrs=dict(sensor="test_sensor2", mode="L"),
                       dims=["y", "x"])
        e = Enhancer()
        assert e.enhancement_tree is not None
        get_enhanced_image(ds, enhance=e)
        assert (set(pathlib.Path(config) for config in e.sensor_enhancement_configs) ==
                {test_configs_path / self.ENH_FN2,
                 test_configs_path / self.ENH_ENH_FN2})

    def test_no_enhance(self):
        """Test turning off enhancements."""
        from xarray import DataArray

        from satpy.enhancements.enhancer import get_enhanced_image
        ds = DataArray(np.arange(1, 11.).reshape((2, 5)),
                       attrs=dict(name="test1", sensor="test_sensor", mode="L"),
                       dims=["y", "x"])
        img = get_enhanced_image(ds, enhance=False)
        np.testing.assert_allclose(img.data.data.compute().squeeze(), ds.data)

    def test_writer_no_enhance(self):
        """Test turning off enhancements with writer."""
        from xarray import DataArray
        ds = DataArray(np.arange(1, 11.).reshape((2, 5)),
                       attrs=dict(name="test1", sensor="test_sensor", mode="L"),
                       dims=["y", "x"])
        writer = _CustomImageWriter(enhance=False)
        writer.save_datasets((ds,), compute=False)
        img = writer.img
        np.testing.assert_allclose(img.data.data.compute().squeeze(), ds.data)

    def test_writer_custom_enhance(self):
        """Test using custom enhancements with writer."""
        from xarray import DataArray

        from satpy.enhancements.enhancer import Enhancer
        ds = DataArray(np.arange(1, 11.).reshape((2, 5)),
                       attrs=dict(name="test1", sensor="test_sensor", mode="L"),
                       dims=["y", "x"])
        enhance = Enhancer()
        writer = _CustomImageWriter(enhance=enhance)
        writer.save_datasets((ds,), compute=False)
        img = writer.img
        np.testing.assert_almost_equal(img.data.isel(bands=0).max().values, 1.)

    def test_enhance_with_sensor_entry(self, test_configs_path):
        """Test enhancing an image with a configuration section."""
        from xarray import DataArray

        from satpy.enhancements.enhancer import Enhancer, get_enhanced_image
        ds = DataArray(np.arange(1, 11.).reshape((2, 5)),
                       attrs=dict(name="test1", sensor="test_sensor", mode="L"),
                       dims=["y", "x"])
        e = Enhancer()
        assert e.enhancement_tree is not None
        img = get_enhanced_image(ds, enhance=e)
        assert (set(pathlib.Path(config) for config in e.sensor_enhancement_configs) ==
                {test_configs_path / self.ENH_FN,
                 test_configs_path / self.ENH_ENH_FN})
        np.testing.assert_almost_equal(img.data.isel(bands=0).max().values,
                                       1.)

        ds = DataArray(da.arange(1, 11., chunks=5).reshape((2, 5)),
                       attrs=dict(name="test1", sensor="test_sensor", mode="L"),
                       dims=["y", "x"])
        e = Enhancer()
        assert e.enhancement_tree is not None
        img = get_enhanced_image(ds, enhance=e)
        assert (set(pathlib.Path(config) for config in e.sensor_enhancement_configs) ==
                {test_configs_path / self.ENH_FN,
                 test_configs_path / self.ENH_ENH_FN})
        np.testing.assert_almost_equal(img.data.isel(bands=0).max().values, 1.)

    def test_enhance_with_sensor_entry2(self, test_configs_path):
        """Test enhancing an image with a more detailed configuration section."""
        from xarray import DataArray

        from satpy.enhancements.enhancer import Enhancer, get_enhanced_image
        ds = DataArray(np.arange(1, 11.).reshape((2, 5)),
                       attrs=dict(name="test1", units="kelvin",
                                  sensor="test_sensor", mode="L"),
                       dims=["y", "x"])
        e = Enhancer()
        assert e.enhancement_tree is not None
        img = get_enhanced_image(ds, enhance=e)
        assert (set(pathlib.Path(config) for config in e.sensor_enhancement_configs) ==
                {test_configs_path / self.ENH_FN,
                 test_configs_path / self.ENH_ENH_FN})
        np.testing.assert_almost_equal(img.data.isel(bands=0).max().values, 0.5)


class TestReaderEnhancerConfigs(_BaseCustomEnhancementConfigTests):
    """Test enhancement configs that use reader name."""

    ENH_FN = "test_sensor1.yaml"

    # NOTE: The sections are ordered in a special way so that if 'reader' key
    #   isn't provided that we'll get the section we didn't want and all tests
    #   will fail. Otherwise the correct sections get chosen just by the order
    #   of how they are added to the decision tree.
    TEST_CONFIGS = {
        ENH_FN: """
enhancements:
  default_reader2:
    reader: reader2
    operations:
    - name: stretch
      method: !!python/name:satpy.enhancements.contrast.stretch
      kwargs: {stretch: crude, min_stretch: 0, max_stretch: 75}
  default:
    operations:
    - name: stretch
      method: !!python/name:satpy.enhancements.contrast.stretch
      kwargs: {stretch: crude, min_stretch: 0, max_stretch: 100}
  test1_reader2_specific:
    name: test1
    reader: reader2
    operations:
    - name: stretch
      method: !!python/name:satpy.enhancements.contrast.stretch
      kwargs: {stretch: crude, min_stretch: 0, max_stretch: 50}
  test1_reader1_specific:
    name: test1
    reader: reader1
    operations:
    - name: stretch
      method: !!python/name:satpy.enhancements.contrast.stretch
      kwargs: {stretch: crude, min_stretch: 0, max_stretch: 200}
            """,
    }

    def _get_test_data_array(self):
        from xarray import DataArray
        ds = DataArray(np.arange(1, 11.).reshape((2, 5)),
                       attrs={
                           "name": "test1",
                           "sensor": "test_sensor1",
                           "mode": "L",
                       },
                       dims=["y", "x"])
        return ds

    def _get_enhanced_image(self, data_arr, test_configs_path):
        from satpy.enhancements.enhancer import Enhancer, get_enhanced_image
        e = Enhancer()
        assert e.enhancement_tree is not None
        img = get_enhanced_image(data_arr, enhance=e)
        # make sure that both configs were loaded
        assert (set(pathlib.Path(config) for config in e.sensor_enhancement_configs) ==
                {test_configs_path / self.ENH_FN})
        return img

    def test_no_reader(self, test_configs_path):
        """Test that a DataArray with no 'reader' metadata works."""
        data_arr = self._get_test_data_array()
        img = self._get_enhanced_image(data_arr, test_configs_path)
        # no reader available, should use default no specified reader
        np.testing.assert_allclose(img.data.values[0], data_arr.data / 100.0)

    def test_no_matching_reader(self, test_configs_path):
        """Test that a DataArray with no matching 'reader' works."""
        data_arr = self._get_test_data_array()
        data_arr.attrs["reader"] = "reader3"
        img = self._get_enhanced_image(data_arr, test_configs_path)
        # no reader available, should use default no specified reader
        np.testing.assert_allclose(img.data.values[0], data_arr.data / 100.0)

    def test_only_reader_matches(self, test_configs_path):
        """Test that a DataArray with only a matching 'reader' works."""
        data_arr = self._get_test_data_array()
        data_arr.attrs["reader"] = "reader2"
        data_arr.attrs["name"] = "not_configured"
        img = self._get_enhanced_image(data_arr, test_configs_path)
        # no reader available, should use default no specified reader
        np.testing.assert_allclose(img.data.values[0], data_arr.data / 75.0)

    def test_reader_and_name_match(self, test_configs_path):
        """Test that a DataArray with a matching 'reader' and 'name' works."""
        data_arr = self._get_test_data_array()
        data_arr.attrs["reader"] = "reader2"
        img = self._get_enhanced_image(data_arr, test_configs_path)
        # no reader available, should use default no specified reader
        np.testing.assert_allclose(img.data.values[0], data_arr.data / 50.0)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2018-2022 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for AGRI compositors."""

import unittest


class TestAGRIComposites(unittest.TestCase):
    """Test AGRI-specific composites."""

    def test_load_composite_yaml(self):
        """Test loading the yaml for this sensor."""
        from satpy.composites.config_loader import load_compositor_configs_for_sensors
        load_compositor_configs_for_sensors(["agri"])

    def test_simulated_red(self):
        """Test creating a fake 'red' band."""
        import dask.array as da
        import numpy as np
        import xarray as xr
        from pyresample.geometry import AreaDefinition

        from satpy.composites.agri import SimulatedRed
        rows = 5
        cols = 10
        area = AreaDefinition(
            "test", "test", "test",
            {"proj": "eqc", "lon_0": 0.0,
             "lat_0": 0.0},
            cols, rows,
            (-20037508.34, -10018754.17, 20037508.34, 10018754.17))

        comp = SimulatedRed("red", prerequisites=("C01", "C02", "C03"),
                            standard_name="toa_bidirectional_reflectance")
        c01 = xr.DataArray(da.zeros((rows, cols), chunks=25) + 0.25,
                           dims=("y", "x"),
                           attrs={"name": "C01", "area": area})
        c02 = xr.DataArray(da.zeros((rows, cols), chunks=25) + 0.30,
                           dims=("y", "x"),
                           attrs={"name": "C02", "area": area})
        res = comp((c01, c02))
        assert isinstance(res, xr.DataArray)
        assert isinstance(res.data, da.Array)
        assert res.attrs["name"] == "red"
        assert res.attrs["standard_name"] == "toa_bidirectional_reflectance"
        data = res.compute()
        np.testing.assert_allclose(data, 0.24252874)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Nowcasting SAF common PPS&MSG NetCDF/CF format reader.

References:
   - The NWCSAF GEO 2018 products documentation: http://www.nwcsaf.org/web/guest/archive

"""

import datetime as dt
import functools
import logging
import os
from contextlib import suppress

import dask.array as da
import numpy as np
import xarray as xr
from pyproj import CRS
from pyresample.geometry import AreaDefinition

from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.readers.core.utils import unzip_file
from satpy.utils import get_chunk_size_limit

logger = logging.getLogger(__name__)

CHUNK_SIZE = get_chunk_size_limit()

SENSOR = {
    "NOAA-19": "avhrr-3",
    "NOAA-18": "avhrr-3",
    "NOAA-15": "avhrr-3",
    "Metop-A": "avhrr-3",
    "Metop-B": "avhrr-3",
    "Metop-C": "avhrr-3",
    "EOS-Aqua": "modis",
    "EOS-Terra": "modis",
    "Suomi-NPP": "viirs",
    "NOAA-20": "viirs",
    "NOAA-21": "viirs",
    "NOAA-22": "viirs",
    "NOAA-23": "viirs",
    "JPSS-1": "viirs",
    "Metop-SG-A1": "metimage",
    "Metop-SG-A2": "metimage",
    "Metop-SG-A3": "metimage",
    "GOES-16": "abi",
    "GOES-17": "abi",
    "Himawari-8": "ahi",
    "Himawari-9": "ahi",
    "Meteosat-12": "fci",
}


PLATFORM_NAMES = {
    "MSG1": "Meteosat-8",
    "MSG2": "Meteosat-9",
    "MSG3": "Meteosat-10",
    "MSG4": "Meteosat-11",
    "MTI1": "Meteosat-12",
    "GOES16": "GOES-16",
    "GOES17": "GOES-17",
}


class NcNWCSAF(BaseFileHandler):
    """NWCSAF PPS&MSG NetCDF reader."""

    def __init__(self, filename, filename_info, filetype_info):
        """Init method."""
        super(NcNWCSAF, self).__init__(filename, filename_info,
                                       filetype_info)

        self._unzipped = unzip_file(self.filename)
        if self._unzipped:
            self.filename = self._unzipped

        self.cache = {}
        self.nc = xr.open_dataset(self.filename,
                                  decode_cf=True,
                                  mask_and_scale=False,
                                  chunks=CHUNK_SIZE)

        self.nc = self.nc.rename({"nx": "x", "ny": "y"})
        self.sw_version = self.nc.attrs["source"]

        self.pps = False
        self.platform_name = None
        self.sensor = None
        self.file_key_prefix = filetype_info.get("file_key_prefix", "")

        try:
            # NWCSAF/Geo:
            try:
                kwrgs = {"sat_id": self.nc.attrs["satellite_identifier"]}
            except KeyError:
                kwrgs = {"sat_id": self.nc.attrs["satellite_identifier"].astype(str)}
        except KeyError:
            # NWCSAF/PPS:
            kwrgs = {"platform_name": self.nc.attrs["platform"]}

        self.set_platform_and_sensor(**kwrgs)

        self.upsample_geolocation = functools.lru_cache(maxsize=1)(
            self._upsample_geolocation_uncached
        )

    def set_platform_and_sensor(self, **kwargs):
        """Set some metadata: platform_name, sensors, and pps (identifying PPS or Geo)."""
        try:
            # NWCSAF/Geo
            self.platform_name = PLATFORM_NAMES.get(kwargs["sat_id"], kwargs["sat_id"])
        except KeyError:
            # NWCSAF/PPS
            self.platform_name = kwargs["platform_name"]
            self.pps = True

        self.sensor = set([SENSOR.get(self.platform_name, "seviri")])

    def remove_timedim(self, var):
        """Remove time dimension from dataset."""
        if var.dims[0] == "time":
            data = var[0, :, :]
            data.attrs = var.attrs
            var = data
        return var

    def drop_xycoords(self, variable):
        """Drop x, y coords when y is scan line number."""
        try:
            if variable.coords["y"].attrs["long_name"] == "scan line number":
                return variable.drop_vars(["y", "x"])
        except KeyError:
            pass
        return variable

    def get_dataset(self, dsid, info):
        """Load a dataset."""
        dsid_name = dsid["name"]
        if dsid_name in self.cache:
            logger.debug("Get the data set from cache: %s.", dsid_name)
            return self.cache[dsid_name]
        if dsid_name in ["lon", "lat"] and dsid_name not in self.nc:
            # Get full resolution lon,lat from the reduced (tie points) grid
            lon, lat = self.upsample_geolocation()
            if dsid_name == "lon":
                return lon
            else:
                return lat

        logger.debug("Reading %s.", dsid_name)
        file_key = self._get_filekeys(dsid_name, info)
        variable = self.nc[file_key]
        variable = self.remove_timedim(variable)
        variable = self.scale_dataset(variable, info)
        variable = self.drop_xycoords(variable)

        self.get_orbital_parameters(variable)
        variable.attrs["start_time"] = self.start_time
        variable.attrs["end_time"] = self.end_time

        return variable

    def get_orbital_parameters(self, variable):
        """Get the orbital parameters from the file if possible (geo)."""
        with suppress(KeyError):
            gdal_params = dict(elt.strip("+").split("=") for elt in self.nc.attrs["gdal_projection"].split())
            variable.attrs["orbital_parameters"] = dict(
                satellite_nominal_altitude=float(gdal_params["h"]),
                satellite_nominal_longitude=float(self.nc.attrs["sub-satellite_longitude"]),
                satellite_nominal_latitude=0)

    def _get_varname_in_file(self, info, info_type="file_key"):
        if isinstance(info[info_type], list):
            for key in info[info_type]:
                file_key = self.file_key_prefix + key
                if file_key in self.nc:
                    return file_key
        return self.file_key_prefix + info[info_type]

    def _get_filekeys(self, dsid_name, info):
        try:
            file_key = self._get_varname_in_file(info, info_type="file_key")
        except KeyError:
            file_key = dsid_name
        return file_key

    def scale_dataset(self, variable, info):
        """Scale the data set, applying the attributes from the netCDF file.

        The scale and offset attributes will then be removed from the resulting variable.
        """
        variable = remove_empties(variable)

        scale = variable.attrs.get("scale_factor", np.array(1, dtype=variable.dtype))
        offset = variable.attrs.get("add_offset", np.array(0, dtype=variable.dtype))
        if "_FillValue" in variable.attrs:
            variable.attrs["scaled_FillValue"] = variable.attrs["_FillValue"] * scale + offset
        if np.issubdtype((scale + offset).dtype, np.floating) or np.issubdtype(variable.dtype, np.floating):
            variable = self._mask_variable(variable)
        attrs = variable.attrs.copy()
        variable = variable * scale + offset
        variable.attrs = attrs
        if "valid_range" in variable.attrs:
            variable.attrs["valid_range"] = variable.attrs["valid_range"] * scale + offset

        variable.attrs.pop("add_offset", None)
        variable.attrs.pop("scale_factor", None)

        variable.attrs.update({"platform_name": self.platform_name,
                               "sensor": self.sensor})

        if not variable.attrs.get("standard_name", "").endswith("status_flag"):
            # TODO: do we really need to add units to everything ?
            variable.attrs.setdefault("units", "1")

        ancillary_names = variable.attrs.get("ancillary_variables", "")
        try:
            variable.attrs["ancillary_variables"] = ancillary_names.split()
        except AttributeError:
            pass

        if "palette_meanings" in variable.attrs:
            variable = self._prepare_variable_for_palette(variable, info)

        if "standard_name" in info:
            variable.attrs.setdefault("standard_name", info["standard_name"])
        variable = self._adjust_variable_for_legacy_software(variable)

        return variable

    @staticmethod
    def _mask_variable(variable):
        if "_FillValue" in variable.attrs:
            variable = variable.where(
                variable != variable.attrs["_FillValue"])
            variable.attrs["_FillValue"] = np.nan
        if "valid_range" in variable.attrs:
            variable = variable.where(
                variable <= variable.attrs["valid_range"][1])
            variable = variable.where(
                variable >= variable.attrs["valid_range"][0])
        if "valid_max" in variable.attrs:
            variable = variable.where(
                variable <= variable.attrs["valid_max"])
        if "valid_min" in variable.attrs:
            variable = variable.where(
                variable >= variable.attrs["valid_min"])
        return variable

    def _prepare_variable_for_palette(self, variable, info):
        try:
            so_dataset = self.nc[self._get_varname_in_file(info, info_type="scale_offset_dataset")]
        except KeyError:
            scale = 1
            offset = 0
            fill_value = 255
        else:
            scale = so_dataset.attrs["scale_factor"]
            offset = so_dataset.attrs["add_offset"]
            fill_value = so_dataset.attrs["_FillValue"]
        variable.attrs["palette_meanings"] = [int(val)
                                              for val in variable.attrs["palette_meanings"].split()]

        if fill_value not in variable.attrs["palette_meanings"] and "fill_value_color" in variable.attrs:
            variable.attrs["palette_meanings"] = [fill_value] + variable.attrs["palette_meanings"]
            variable = xr.DataArray(da.vstack((np.array(variable.attrs["fill_value_color"]), variable.data)),
                                    coords=variable.coords, dims=variable.dims, attrs=variable.attrs)
        val, idx = np.unique(variable.attrs["palette_meanings"], return_index=True)
        variable.attrs["palette_meanings"] = val * scale + offset
        variable = variable[idx]
        return variable

    def _adjust_variable_for_legacy_software(self, variable):
        if self.sw_version == "NWC/PPS version v2014" and variable.attrs.get("standard_name") == "cloud_top_altitude":
            # pps 2014 valid range and palette don't match
            variable.attrs["valid_range"] = (0., 9000.)
        if (self.sw_version == "NWC/PPS version v2014" and
                variable.attrs.get("long_name") == "RGB Palette for ctth_alti"):
            # pps 2014 palette has the nodata color (black) first
            variable = variable[1:, :]
        return variable

    def _upsample_geolocation_uncached(self):
        """Upsample the geolocation (lon,lat) from the tiepoint grid."""
        from geotiepoints import SatelliteInterpolator

        # Read the fields needed:
        col_indices = self.nc["nx_reduced"].values
        row_indices = self.nc["ny_reduced"].values
        lat_reduced = self.scale_dataset(self.nc["lat_reduced"], {})
        lon_reduced = self.scale_dataset(self.nc["lon_reduced"], {})

        shape = (self.nc["y"].shape[0], self.nc["x"].shape[0])
        cols_full = np.arange(shape[1])
        rows_full = np.arange(shape[0])

        satint = SatelliteInterpolator((lon_reduced.values, lat_reduced.values),
                                       (row_indices,
                                        col_indices),
                                       (rows_full, cols_full))

        lons, lats = satint.interpolate()
        lon = xr.DataArray(lons, attrs=lon_reduced.attrs, dims=["y", "x"])
        lat = xr.DataArray(lats, attrs=lat_reduced.attrs, dims=["y", "x"])
        lat = self.drop_xycoords(lat)
        lon = self.drop_xycoords(lon)
        return lon, lat

    def get_area_def(self, dsid):
        """Get the area definition of the datasets in the file.

        Only applicable for MSG products!
        """
        if self.pps:
            # PPS:
            raise NotImplementedError

        if dsid["name"].endswith("_pal"):
            raise NotImplementedError

        crs, area_extent = self._get_projection()
        crs, area_extent = self._ensure_crs_extents_in_meters(crs, area_extent)
        variable = self.remove_timedim(self.nc[dsid["name"]])
        nlines, ncols = variable.shape
        area = AreaDefinition("some_area_name",
                              "On-the-fly area",
                              "geosmsg",
                              crs,
                              ncols,
                              nlines,
                              area_extent)

        return area

    @staticmethod
    def _ensure_crs_extents_in_meters(crs, area_extent):
        """Fix units in Earth shape, satellite altitude and 'units' attribute."""
        import warnings
        if "kilo" in crs.axis_info[0].unit_name:
            with warnings.catch_warnings():
                # The proj dict route is the only feasible way to modify the area, suppress the warning it causes
                warnings.filterwarnings("ignore", category=UserWarning,
                                        message="You will likely lose important projection information")
                proj_dict = crs.to_dict()
            proj_dict["units"] = "m"
            if "a" in proj_dict:
                proj_dict["a"] *= 1000.
            if "b" in proj_dict:
                proj_dict["b"] *= 1000.
            if "R" in proj_dict:
                proj_dict["R"] *= 1000.
            proj_dict["h"] *= 1000.
            area_extent = tuple([val * 1000. for val in area_extent])
            crs = CRS.from_dict(proj_dict)
        return crs, area_extent

    def __del__(self):
        """Delete the instance."""
        if self._unzipped:
            try:
                os.remove(self._unzipped)
            except OSError:
                pass

    @property
    def start_time(self):
        """Return the start time of the object."""
        try:
            return read_nwcsaf_time(self.nc.attrs["nominal_product_time"])
        except KeyError:
            return read_nwcsaf_time(self.nc.attrs["time_coverage_start"])

    @property
    def end_time(self):
        """Return the end time of the object."""
        return read_nwcsaf_time(self.nc.attrs["time_coverage_end"])

    @property
    def sensor_names(self):
        """List of sensors represented in this file."""
        return self.sensor

    def _get_projection(self):
        """Get projection from the NetCDF4 attributes."""
        try:
            proj_str = self.nc.attrs["gdal_projection"]
        except TypeError:
            proj_str = self.nc.attrs["gdal_projection"].decode()

        # Check the a/b/h units
        radius_a = proj_str.split("+a=")[-1].split()[0]
        if float(radius_a) > 10e3:
            units = "m"
            scale = 1.0
        else:
            units = "km"
            scale = 1e3

        if "units" not in proj_str:
            proj_str = proj_str + " +units=" + units

        area_extent = (float(self.nc.attrs["gdal_xgeo_up_left"]) / scale,
                       float(self.nc.attrs["gdal_ygeo_low_right"]) / scale,
                       float(self.nc.attrs["gdal_xgeo_low_right"]) / scale,
                       float(self.nc.attrs["gdal_ygeo_up_left"]) / scale)

        crs = CRS.from_string(proj_str)
        return crs, area_extent


def remove_empties(variable):
    """Remove empty objects from the *variable*'s attrs."""
    import h5py
    for key, val in variable.attrs.items():
        if isinstance(val, h5py._hl.base.Empty):
            variable.attrs.pop(key)

    return variable


def read_nwcsaf_time(time_value):
    """Read the time, nwcsaf-style."""
    try:
        # MSG:
        try:
            return dt.datetime.strptime(time_value, "%Y-%m-%dT%H:%M:%SZ")
        except TypeError:  # Remove this in summer 2024 (this is not needed since h5netcdf 0.14)
            return dt.datetime.strptime(time_value.astype(str), "%Y-%m-%dT%H:%M:%SZ")
    except ValueError:
        # PPS:
        return dt.datetime.strptime(time_value, "%Y%m%dT%H%M%S%fZ")



# Copyright (c) 2015-2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Helpers for adding overlays and decorations to images."""
from __future__ import annotations

import logging
import warnings

import numpy as np
import xarray as xr
from dask import array as da

from satpy.utils import get_legacy_chunk_size

LOG = logging.getLogger(__name__)
CHUNK_SIZE = get_legacy_chunk_size()


def _burn_overlay(img, image_metadata, area, cw_, overlays):
    """Burn the overlay in the image array."""
    del image_metadata
    cw_.add_overlay_from_dict(overlays, area, background=img)
    return img


def add_overlay(orig_img, area, coast_dir, color=None, width=None, resolution=None,
                level_coast=None, level_borders=None, fill_value=None,
                grid=None, overlays=None):
    """Add coastline, political borders and grid(graticules) to image.

    Uses ``color`` for feature colors where ``color`` is a 3-element tuple
    of integers between 0 and 255 representing (R, G, B).

    .. warning::

        This function currently loses the data mask (alpha band).

    ``resolution`` is chosen automatically if None (default),
    otherwise it should be one of:

    +-----+-------------------------+---------+
    | 'f' | Full resolution         | 0.04 km |
    +-----+-------------------------+---------+
    | 'h' | High resolution         | 0.2 km  |
    +-----+-------------------------+---------+
    | 'i' | Intermediate resolution | 1.0 km  |
    +-----+-------------------------+---------+
    | 'l' | Low resolution          | 5.0 km  |
    +-----+-------------------------+---------+
    | 'c' | Crude resolution        | 25  km  |
    +-----+-------------------------+---------+

    ``grid`` is a dictionary with key values as documented in detail in pycoast

    eg. overlay={'grid': {'major_lonlat': (10, 10),
                          'write_text': False,
                          'outline': (224, 224, 224),
                          'width': 0.5}}

    Here major_lonlat is plotted every 10 deg for both longitude and latitude,
    no labels for the grid lines are plotted, the color used for the grid lines
    is light gray, and the width of the gratucules is 0.5 pixels.

    For grid if aggdraw is used, font option is mandatory, if not
    ``write_text`` is set to False::

        font = aggdraw.Font('black', '/usr/share/fonts/truetype/msttcorefonts/Arial.ttf',
                            opacity=127, size=16)

    """
    if area is None:
        raise ValueError("Area of image is None, can't add overlay.")

    from pycoast import ContourWriterAGG
    from pyresample.area_config import get_area_def

    if isinstance(area, str):
        area = get_area_def(area)
    LOG.info("Add coastlines and political borders to image.")

    old_args = [color, width, resolution, grid, level_coast, level_borders]
    if any(arg is not None for arg in old_args):
        warnings.warn(
            "'color', 'width', 'resolution', 'grid', 'level_coast', 'level_borders'"
            " arguments will be deprecated soon. Please use 'overlays' instead.",
            DeprecationWarning,
            stacklevel=2
        )
    if hasattr(orig_img, "convert"):
        # image must be in RGB space to work with pycoast/pydecorate
        res_mode = ("RGBA" if orig_img.final_mode(fill_value).endswith("A") else "RGB")
        orig_img = orig_img.convert(res_mode)
    elif not orig_img.mode.startswith("RGB"):
        raise RuntimeError("'trollimage' 1.6+ required to support adding "
                           "overlays/decorations to non-RGB data.")

    if overlays is None:
        overlays = _create_overlays_dict(color, width, grid, level_coast, level_borders)

    cw_ = ContourWriterAGG(coast_dir)
    new_image = orig_img.apply_pil(_burn_overlay, res_mode,
                                   None, {"fill_value": fill_value},
                                   (area, cw_, overlays), None)
    return new_image


def _create_overlays_dict(color, width, grid, level_coast, level_borders):
    """Fill in the overlays dict."""
    overlays = dict()
    # fill with sensible defaults
    general_params = {"outline": color or (0, 0, 0),
                      "width": width or 0.5}
    for key, val in general_params.items():
        if val is not None:
            overlays.setdefault("coasts", {}).setdefault(key, val)
            overlays.setdefault("borders", {}).setdefault(key, val)
    if level_coast is None:
        level_coast = 1
    overlays.setdefault("coasts", {}).setdefault("level", level_coast)
    if level_borders is None:
        level_borders = 1
    overlays.setdefault("borders", {}).setdefault("level", level_borders)
    if grid is not None:
        if "major_lonlat" in grid and grid["major_lonlat"]:
            major_lonlat = grid.pop("major_lonlat")
            minor_lonlat = grid.pop("minor_lonlat", major_lonlat)
            grid.update({"Dlonlat": major_lonlat, "dlonlat": minor_lonlat})
        for key, val in grid.items():
            overlays.setdefault("grid", {}).setdefault(key, val)
    return overlays


def add_text(orig, dc, img, text):
    """Add text to an image using the pydecorate package.

    All the features of pydecorate's ``add_text`` are available.
    See documentation of :doc:`pydecorate:index` for more info.

    """
    from trollimage.xrimage import XRImage

    LOG.info("Add text to image.")

    dc.add_text(**text)

    arr = da.from_array(np.array(img) / 255.0, chunks=CHUNK_SIZE)

    new_data = xr.DataArray(arr, dims=["y", "x", "bands"],
                            coords={"y": orig.data.coords["y"],
                                    "x": orig.data.coords["x"],
                                    "bands": list(img.mode)},
                            attrs=orig.data.attrs)
    return XRImage(new_data)


def add_logo(orig, dc, img, logo):
    """Add logos or other images to an image using the pydecorate package.

    All the features of pydecorate's ``add_logo`` are available.
    See documentation of :doc:`pydecorate:index` for more info.

    """
    from trollimage.xrimage import XRImage

    LOG.info("Add logo to image.")

    dc.add_logo(**logo)

    arr = da.from_array(np.array(img) / 255.0, chunks=CHUNK_SIZE)

    new_data = xr.DataArray(arr, dims=["y", "x", "bands"],
                            coords={"y": orig.data.coords["y"],
                                    "x": orig.data.coords["x"],
                                    "bands": list(img.mode)},
                            attrs=orig.data.attrs)
    return XRImage(new_data)


def add_scale(orig, dc, img, scale):
    """Add scale to an image using the pydecorate package.

    All the features of pydecorate's ``add_scale`` are available.
    See documentation of :doc:`pydecorate:index` for more info.

    """
    from trollimage.xrimage import XRImage

    LOG.info("Add scale to image.")

    dc.add_scale(**scale)

    arr = da.from_array(np.array(img) / 255.0, chunks=CHUNK_SIZE)

    new_data = xr.DataArray(arr, dims=["y", "x", "bands"],
                            coords={"y": orig.data.coords["y"],
                                    "x": orig.data.coords["x"],
                                    "bands": list(img.mode)},
                            attrs=orig.data.attrs)
    return XRImage(new_data)


def add_decorate(orig, fill_value=None, **decorate):
    """Decorate an image with text and/or logos/images.

    This call adds text/logos in order as given in the input to keep the
    alignment features available in pydecorate.

    An example of the decorate config::

        decorate = {
            'decorate': [
                {'logo': {'logo_path': <path to a logo>, 'height': 143, 'bg': 'white', 'bg_opacity': 255}},
                {'text': {'txt': start_time_txt,
                          'align': {'top_bottom': 'bottom', 'left_right': 'right'},
                          'font': <path to ttf font>,
                          'font_size': 22,
                          'height': 30,
                          'bg': 'black',
                          'bg_opacity': 255,
                          'line': 'white'}}
            ]
        }

    Any numbers of text/logo in any order can be added to the decorate list,
    but the order of the list is kept as described above.

    Note that a feature given in one element, eg. bg (which is the background color)
    will also apply on the next elements  unless a new value is given.

    align is a special keyword telling where in the image to start adding features, top_bottom is either top or bottom
    and left_right is either left or right.
    """
    LOG.info("Decorate image.")

    # Need to create this here to possible keep the alignment
    # when adding text and/or logo with pydecorate
    if hasattr(orig, "convert"):
        # image must be in RGB space to work with pycoast/pydecorate
        orig = orig.convert("RGBA" if orig.mode.endswith("A") else "RGB")
    elif not orig.mode.startswith("RGB"):
        raise RuntimeError("'trollimage' 1.6+ required to support adding "
                           "overlays/decorations to non-RGB data.")
    img_orig = orig.pil_image(fill_value=fill_value)
    from pydecorate import DecoratorAGG
    dc = DecoratorAGG(img_orig)

    # decorate need to be a list to maintain the alignment
    # as ordered in the list
    img = orig
    if "decorate" in decorate:
        for dec in decorate["decorate"]:
            if "logo" in dec:
                img = add_logo(img, dc, img_orig, logo=dec["logo"])
            elif "text" in dec:
                img = add_text(img, dc, img_orig, text=dec["text"])
            elif "scale" in dec:
                img = add_scale(img, dc, img_orig, scale=dec["scale"])
    return img



# Copyright (c) 2015-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Sharpening compositors."""

from __future__ import annotations

import logging

import dask.array as da
import numpy as np
import xarray as xr

from .core import GenericCompositor, IncompatibleAreas, enhance2dataset

LOG = logging.getLogger(__name__)


class RatioSharpenedRGB(GenericCompositor):
    """Sharpen RGB bands with ratio of a high resolution band to a lower resolution version.

    Any pixels where the ratio is computed to be negative or infinity, it is
    reset to 1. Additionally, the ratio is limited to 1.5 on the high end to
    avoid high changes due to small discrepancies in instrument detector
    footprint. Note that the input data to this compositor must already be
    resampled so all data arrays are the same shape.

    Example::

        R_lo -  1000m resolution - shape=(2000, 2000)
        G - 1000m resolution - shape=(2000, 2000)
        B - 1000m resolution - shape=(2000, 2000)
        R_hi -  500m resolution - shape=(4000, 4000)

        ratio = R_hi / R_lo
        new_R = R_hi
        new_G = G * ratio
        new_B = B * ratio

    In some cases, there could be multiple high resolution bands::

        R_lo -  1000m resolution - shape=(2000, 2000)
        G_hi - 500m resolution - shape=(4000, 4000)
        B - 1000m resolution - shape=(2000, 2000)
        R_hi -  500m resolution - shape=(4000, 4000)

    To avoid the green band getting involved in calculating ratio or sharpening,
    add "neutral_resolution_band: green" in the YAML config file. This way
    only the blue band will get sharpened::

        ratio = R_hi / R_lo
        new_R = R_hi
        new_G = G_hi
        new_B = B * ratio

    """

    def __init__(self, *args, **kwargs):
        """Instanciate the ration sharpener."""
        self.high_resolution_color = kwargs.pop("high_resolution_band", "red")
        self.neutral_resolution_color = kwargs.pop("neutral_resolution_band", None)
        if self.high_resolution_color not in ["red", "green", "blue", None]:
            raise ValueError("RatioSharpenedRGB.high_resolution_band must "
                             "be one of ['red', 'green', 'blue', None]. Not "
                             "'{}'".format(self.high_resolution_color))
        if self.neutral_resolution_color not in ["red", "green", "blue", None]:
            raise ValueError("RatioSharpenedRGB.neutral_resolution_band must "
                             "be one of ['red', 'green', 'blue', None]. Not "
                             "'{}'".format(self.neutral_resolution_color))
        super(RatioSharpenedRGB, self).__init__(*args, **kwargs)

    def __call__(self, datasets, optional_datasets=None, **info):
        """Sharpen low resolution datasets by multiplying by the ratio of ``high_res / low_res``.

        The resulting RGB has the units attribute removed.
        """
        if len(datasets) != 3:
            raise ValueError("Expected 3 datasets, got %d" % (len(datasets), ))
        if not all(x.shape == datasets[0].shape for x in datasets[1:]) or \
                (optional_datasets and
                 optional_datasets[0].shape != datasets[0].shape):
            raise IncompatibleAreas("RatioSharpening requires datasets of "
                                    "the same size. Must resample first.")

        optional_datasets = tuple() if optional_datasets is None else optional_datasets
        datasets = self.match_data_arrays(datasets + optional_datasets)
        red, green, blue, new_attrs = self._get_and_sharpen_rgb_data_arrays_and_meta(datasets, optional_datasets)
        combined_info = self._combined_sharpened_info(info, new_attrs)
        res = super(RatioSharpenedRGB, self).__call__((red, green, blue,), **combined_info)
        res.attrs.pop("units", None)
        return res

    def _get_and_sharpen_rgb_data_arrays_and_meta(self, datasets, optional_datasets):
        new_attrs = {}
        low_res_red = datasets[0]
        low_res_green = datasets[1]
        low_res_blue = datasets[2]
        if optional_datasets and self.high_resolution_color is not None:
            LOG.debug("Sharpening image with high resolution {} band".format(self.high_resolution_color))
            high_res = datasets[3]
            if "rows_per_scan" in high_res.attrs:
                new_attrs.setdefault("rows_per_scan", high_res.attrs["rows_per_scan"])
            new_attrs.setdefault("resolution", high_res.attrs["resolution"])

        else:
            LOG.debug("No sharpening band specified for ratio sharpening")
            high_res = None

        bands = {"red": low_res_red, "green": low_res_green, "blue": low_res_blue}
        if high_res is not None:
            self._sharpen_bands_with_high_res(bands, high_res)

        return bands["red"], bands["green"], bands["blue"], new_attrs

    def _sharpen_bands_with_high_res(self, bands, high_res):
        ratio = da.map_blocks(
            _get_sharpening_ratio,
            high_res.data,
            bands[self.high_resolution_color].data,
            meta=np.array((), dtype=high_res.dtype),
            dtype=high_res.dtype,
            chunks=high_res.chunks,
        )

        bands[self.high_resolution_color] = high_res

        with xr.set_options(keep_attrs=True):
            for color in bands.keys():
                if color != self.neutral_resolution_color and color != self.high_resolution_color:
                    bands[color] = bands[color] * ratio

    def _combined_sharpened_info(self, info, new_attrs):
        combined_info = {}
        combined_info.update(info)
        combined_info.update(new_attrs)
        # Update that information with configured information (including name)
        combined_info.update(self.attrs)
        # Force certain pieces of metadata that we *know* to be true
        combined_info.setdefault("standard_name", "true_color")
        return combined_info


def _get_sharpening_ratio(high_res, low_res):
    with np.errstate(divide="ignore"):
        ratio = high_res / low_res
    # make ratio a no-op (multiply by 1) where the ratio is NaN, infinity,
    # or it is negative.
    ratio[~np.isfinite(ratio) | (ratio < 0)] = 1.0
    # we don't need ridiculously high ratios, they just make bright pixels
    np.clip(ratio, 0, 1.5, out=ratio)
    return ratio


def _mean4(data, offset=(0, 0), block_id=None):
    rows, cols = data.shape
    # we assume that the chunks except the first ones are aligned
    if block_id[0] == 0:
        row_offset = offset[0] % 2
    else:
        row_offset = 0
    if block_id[1] == 0:
        col_offset = offset[1] % 2
    else:
        col_offset = 0
    row_after = (row_offset + rows) % 2
    col_after = (col_offset + cols) % 2
    pad = ((row_offset, row_after), (col_offset, col_after))

    rows2 = rows + row_offset + row_after
    cols2 = cols + col_offset + col_after

    av_data = np.pad(data, pad, "edge")
    new_shape = (int(rows2 / 2.), 2, int(cols2 / 2.), 2)
    with np.errstate(invalid="ignore"):
        data_mean = np.nanmean(av_data.reshape(new_shape), axis=(1, 3))
    data_mean = np.repeat(np.repeat(data_mean, 2, axis=0), 2, axis=1)
    data_mean = data_mean[row_offset:row_offset + rows, col_offset:col_offset + cols]
    return data_mean


class SelfSharpenedRGB(RatioSharpenedRGB):
    """Sharpen RGB with ratio of a band with a strided-version of itself.

    Example::

        R -  500m resolution - shape=(4000, 4000)
        G - 1000m resolution - shape=(2000, 2000)
        B - 1000m resolution - shape=(2000, 2000)

        ratio = R / four_element_average(R)
        new_R = R
        new_G = G * ratio
        new_B = B * ratio

    """

    @staticmethod
    def four_element_average_dask(d):
        """Average every 4 elements (2x2) in a 2D array."""
        try:
            offset = d.attrs["area"].crop_offset
        except (KeyError, AttributeError):
            offset = (0, 0)

        res = d.data.map_blocks(_mean4, offset=offset, dtype=d.dtype, meta=np.ndarray((), dtype=d.dtype))
        return xr.DataArray(res, attrs=d.attrs, dims=d.dims, coords=d.coords)

    def __call__(self, datasets, optional_datasets=None, **attrs):
        """Generate the composite."""
        colors = ["red", "green", "blue"]
        if self.high_resolution_color not in colors:
            raise ValueError("SelfSharpenedRGB requires at least one high resolution band, not "
                             "'{}'".format(self.high_resolution_color))

        high_res = datasets[colors.index(self.high_resolution_color)]
        high_mean = self.four_element_average_dask(high_res)
        red = high_mean if self.high_resolution_color == "red" else datasets[0]
        green = high_mean if self.high_resolution_color == "green" else datasets[1]
        blue = high_mean if self.high_resolution_color == "blue" else datasets[2]
        return super(SelfSharpenedRGB, self).__call__((red, green, blue), optional_datasets=(high_res,), **attrs)


class LuminanceSharpeningCompositor(GenericCompositor):
    """Create a high resolution composite by sharpening a low resolution using high resolution luminance.

    This is done by converting to YCbCr colorspace, replacing Y, and convertin back to RGB.
    """

    def __call__(self, projectables, *args, **kwargs):
        """Generate the composite."""
        from trollimage.image import rgb2ycbcr, ycbcr2rgb
        projectables = self.match_data_arrays(projectables)
        luminance = projectables[0].copy()
        luminance /= 100.
        # Limit between min(luminance) ... 1.0
        luminance = da.where(luminance > 1., 1., luminance)

        # Get the enhanced version of the composite to be sharpened
        rgb_img = enhance2dataset(projectables[1])

        # This all will be eventually replaced with trollimage convert() method
        # ycbcr_img = rgb_img.convert('YCbCr')
        # ycbcr_img.data[0, :, :] = luminance
        # rgb_img = ycbcr_img.convert('RGB')

        # Replace luminance of the IR composite
        y__, cb_, cr_ = rgb2ycbcr(rgb_img.data[0, :, :],
                                  rgb_img.data[1, :, :],
                                  rgb_img.data[2, :, :])

        r__, g__, b__ = ycbcr2rgb(luminance, cb_, cr_)
        y_size, x_size = r__.shape
        r__ = da.reshape(r__, (1, y_size, x_size))
        g__ = da.reshape(g__, (1, y_size, x_size))
        b__ = da.reshape(b__, (1, y_size, x_size))

        rgb_img.data = da.vstack((r__, g__, b__))
        return super(LuminanceSharpeningCompositor, self).__call__(rgb_img, *args, **kwargs)


class SandwichCompositor(GenericCompositor):
    """Make a sandwich product."""

    def __call__(self, projectables, *args, **kwargs):
        """Generate the composite."""
        projectables = self.match_data_arrays(projectables)
        luminance = projectables[0]
        luminance = luminance / 100.
        # Limit between min(luminance) ... 1.0
        luminance = luminance.clip(max=1.)

        # Get the enhanced version of the RGB composite to be sharpened
        rgb_img = enhance2dataset(projectables[1])
        # Ignore alpha band when applying luminance
        rgb_img = rgb_img.where(rgb_img.bands == "A", rgb_img * luminance)
        return super(SandwichCompositor, self).__call__(rgb_img, *args, **kwargs)



# Copyright (c) 2009-2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Reading VIIRS VGAC data."""

import datetime as dt
import logging

import numpy as np
import xarray as xr

from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.utils import get_legacy_chunk_size

CHUNK_SIZE = get_legacy_chunk_size()
logger = logging.getLogger(__name__)


class VGACFileHandler(BaseFileHandler):
    """Reader VGAC data."""

    def __init__(self, filename, filename_info, filetype_info):
        """Init the file handler."""
        super(VGACFileHandler, self).__init__(
            filename, filename_info, filetype_info)

        self.engine = "h5netcdf"
        self._start_time = filename_info["start_time"]
        self._end_time = None
        self.sensor = "viirs"
        self.filename_info = filename_info

    def calibrate(self, data, yaml_info, file_key, nc):
        """Calibrate data."""
        scale_factor = yaml_info.get("scale_factor_nc", 0.0002)
        if file_key + "_LUT" in nc:
            bt_lut = nc[file_key + "_LUT"]
            data = self.convert_to_bt(data, bt_lut, scale_factor)
        if data.attrs["units"] == "percent":
            # Should be removed with later versions of data
            data = self.fix_radiances_not_in_percent(data)
        return data

    def convert_to_bt(self, data, data_lut, scale_factor):
        """Convert radances to brightness temperatures."""
        x = np.arange(0, len(data_lut))
        y = data_lut
        scaled_data = data / scale_factor
        brightness_temperatures = xr.DataArray(np.interp(scaled_data, xp=x, fp=y), coords=data.coords, attrs=data.attrs)
        return brightness_temperatures

    def fix_radiances_not_in_percent(self, data):
        """Scale radiances to percent. This was not done in first version of data."""
        return 100 * data

    def set_time_attrs(self, data):
        """Set time from attributes."""
        if "StartTime" in data.attrs:
            data.attrs["start_time"] = dt.datetime.strptime(data.attrs["StartTime"], "%Y-%m-%dT%H:%M:%S")
            data.attrs["end_time"] = dt.datetime.strptime(data.attrs["EndTime"], "%Y-%m-%dT%H:%M:%S")
            self._end_time = data.attrs["end_time"]
            self._start_time = data.attrs["start_time"]

    def extract_time_data(self, data, nc):
        """Decode time data."""
        reference_time = np.datetime64(dt.datetime.strptime(nc["proj_time0"].attrs["units"],
                                                         "days since %d/%m/%YT%H:%M:%S"))
        delta_part_of_day, delta_full_days = np.modf(nc["proj_time0"].values)
        delta_full_days = np.timedelta64(delta_full_days.astype(np.int64), "D").astype("timedelta64[us]")
        delta_part_of_day = delta_part_of_day * np.timedelta64(1, "D").astype("timedelta64[us]")
        delta_hours = data.values * np.timedelta64(1, "h").astype("timedelta64[us]")
        time_data = xr.DataArray(reference_time + delta_full_days + delta_part_of_day + delta_hours,
                                 coords=data.coords, attrs={"long_name": "Scanline time"})
        return time_data

    def decode_time_variable(self, data, file_key, nc):
        """Decide if time data should be decoded."""
        if file_key != "time":
            return data
        if data.attrs["units"] == "hours since proj_time0":
            return self.extract_time_data(data, nc)
        else:
            raise AttributeError('Unit of time variable in VGAC nc file is not "hours since proj_time0"')

    def get_dataset(self, key, yaml_info):
        """Get dataset."""
        logger.debug("Getting data for: %s", yaml_info["name"])
        nc = xr.open_dataset(self.filename, engine=self.engine, decode_times=False,
                             chunks={"y": CHUNK_SIZE, "x": 800})
        name = yaml_info.get("nc_store_name", yaml_info["name"])
        file_key = yaml_info.get("nc_key", name)
        data = nc[file_key]
        data = self.calibrate(data, yaml_info, file_key, nc)
        data = self.decode_time_variable(data, file_key, nc)
        data.attrs.update(nc.attrs)  # For now add global attributes to all datasets
        data.attrs.update(yaml_info)
        self.set_time_attrs(data)
        return data

    @property
    def start_time(self):
        """Get the start time."""
        return self._start_time

    @property
    def end_time(self):
        """Get the end time."""
        return self._end_time



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2019 Satpy developers
#
# This file is part of Satpy.
#
# Satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Module for testing the satpy.readers.mirs module."""

from __future__ import annotations

import datetime as dt
import os
from unittest import mock

import numpy as np
import pytest
import xarray as xr

from satpy._config import config_search_paths
from satpy.dataset import DataID
from satpy.readers.core.loading import load_reader
from satpy.readers.core.yaml_reader import FileYAMLReader
from satpy.tests.utils import RANDOM_GEN

METOP_FILE = "IMG_SX.M2.D17037.S1601.E1607.B0000001.WE.HR.ORB.nc"
NPP_MIRS_L2_SWATH = "NPR-MIRS-IMG_v11r6_npp_s201702061601000_e201702061607000_c202012201658410.nc"
N20_MIRS_L2_SWATH = "NPR-MIRS-IMG_v11r4_n20_s201702061601000_e201702061607000_c202012201658410.nc"
N21_MIRS_L2_SWATH = "NPR-MIRS-IMG_v11r4_n21_s201702061601000_e201702061607000_c202012201658410.nc"
OTHER_MIRS_L2_SWATH = "NPR-MIRS-IMG_v11r4_gpm_s201702061601000_e201702061607000_c202010080001310.nc"

EXAMPLE_FILES = [METOP_FILE, NPP_MIRS_L2_SWATH, OTHER_MIRS_L2_SWATH]

N_CHANNEL = 22
N_FOV = 96
N_SCANLINE = 100
DEFAULT_FILE_DTYPE = np.float32
DEFAULT_2D_SHAPE = (N_SCANLINE, N_FOV)
DEFAULT_DATE = dt.datetime(2019, 6, 19, 13, 0)
DEFAULT_LAT = np.linspace(23.09356, 36.42844, N_SCANLINE * N_FOV,
                          dtype=DEFAULT_FILE_DTYPE)
DEFAULT_LON = np.linspace(127.6879, 144.5284, N_SCANLINE * N_FOV,
                          dtype=DEFAULT_FILE_DTYPE)
FREQ = xr.DataArray(
    np.array([23.8, 31.4, 50.3, 51.76, 52.8, 53.596, 54.4, 54.94, 55.5,
              57.29, 57.29, 57.29, 57.29, 57.29, 57.29, 88.2, 165.5,
              183.31, 183.31, 183.31, 183.31, 183.31][:N_CHANNEL], dtype=np.float32),
    dims="Channel",
    attrs={"description": "Central Frequencies (GHz)"},
)
POLO = xr.DataArray(
    np.array([2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 3, 3, 3, 3, 3, 3][:N_CHANNEL], dtype=np.int16),
    dims="Channel",
    attrs={"description": "Polarizations"},
)

DS_IDS = ["RR", "longitude", "latitude"]
TEST_VARS = ["btemp_88v", "btemp_165h",
             "btemp_23v", "RR", "Sfc_type"]
DEFAULT_UNITS = {"btemp_88v": "K", "btemp_165h": "K",
                 "btemp_23v": "K", "RR": "mm/hr", "Sfc_type": "1"}
PLATFORM = {"M2": "metop-a", "NPP": "npp", "GPM": "gpm"}
SENSOR = {"m2": "amsu-mhs", "npp": "atms", "gpm": "GPI"}

START_TIME = dt.datetime(2017, 2, 6, 16, 1, 0)
END_TIME = dt.datetime(2017, 2, 6, 16, 7, 0)


def fake_coeff_from_fn(fn):
    """Create Fake Coefficients."""
    ameans = RANDOM_GEN.uniform(261, 267, N_CHANNEL)
    locations = [
        [1, 2],
        [1, 2],
        [3, 4, 5],
        [3, 4, 5],
        [4, 5, 6],
        [5, 6, 7],
        [6, 7, 8],
        [7, 8],
        [9, 10, 11],
        [10, 11],
        [10, 11, 12],
        [11, 12, 13],
        [12, 13],
        [12, 13, 14],
        [14, 15],
        [1, 16],
        [17, 18],
        [18, 19],
        [18, 19, 20],
        [19, 20, 21],
        [20, 21, 22],
        [21, 22],
    ]
    all_nchx = [len(loc) for loc in locations]

    coeff_str = []
    for idx in range(1, N_CHANNEL + 1):
        nx = idx - 1
        coeff_str.append("\n")
        next_line = "   {}  {} {}\n".format(idx, all_nchx[nx], ameans[nx])
        coeff_str.append(next_line)
        next_line = "   {}\n".format("   ".join([str(x) for x in locations[idx - 1]]))
        coeff_str.append(next_line)
        for fov in range(1, N_FOV+1):
            random_coeff = np.ones(all_nchx[nx])
            str_coeff = "  ".join([str(x) for x in random_coeff])
            random_means = np.zeros(all_nchx[nx])
            str_means = " ".join([str(x) for x in random_means])
            error_val = RANDOM_GEN.uniform(0, 4)
            coeffs_line = " {:>2} {:>2}  {} {}   {}\n".format(idx, fov,
                                                              str_coeff,
                                                              str_means,
                                                              error_val)
            coeff_str.append(coeffs_line)

    return coeff_str


def _get_datasets_with_attributes(**kwargs):
    """Represent files with two resolution of variables in them (ex. OCEAN)."""
    bt = xr.DataArray(np.linspace(1830, 3930, N_SCANLINE * N_FOV * N_CHANNEL, dtype=np.int16).
                      reshape(N_SCANLINE, N_FOV, N_CHANNEL),
                      attrs={"long_name": "Channel Temperature (K)",
                             "units": "Kelvin",
                             "coordinates": "Longitude Latitude Freq",
                             "scale_factor": 0.01,
                             "_FillValue": -999,
                             "valid_range": [0, 50000]},
                      dims=("Scanline", "Field_of_view", "Channel"))
    rr = xr.DataArray(RANDOM_GEN.integers(100, 500, size=(N_SCANLINE, N_FOV), dtype=np.int16),
                      attrs={"long_name": "Rain Rate (mm/hr)",
                             "units": "mm/hr",
                             "coordinates": "Longitude Latitude",
                             "scale_factor": 0.1,
                             "_FillValue": -999,
                             "valid_range": [0, 1000]},
                      dims=("Scanline", "Field_of_view"))
    sfc_type = xr.DataArray(RANDOM_GEN.integers(0, 4, size=(N_SCANLINE, N_FOV), dtype=np.int16),
                            attrs={"description": "type of surface:0-ocean," +
                                                  "1-sea ice,2-land,3-snow",
                                   "units": "1",
                                   "coordinates": "Longitude Latitude",
                                   "_FillValue": -999,
                                   "valid_range": [0, 3]
                                   },
                            dims=("Scanline", "Field_of_view"))
    latitude = xr.DataArray(DEFAULT_LAT.reshape(DEFAULT_2D_SHAPE),
                            attrs={"long_name":
                                   "Latitude of the view (-90,90)"},
                            dims=("Scanline", "Field_of_view"))
    longitude = xr.DataArray(DEFAULT_LON.reshape(DEFAULT_2D_SHAPE),
                             attrs={"long_name":
                                    "Longitude of the view (-180,180)"},
                             dims=("Scanline", "Field_of_view"))

    ds_vars = {
        "Freq": FREQ,
        "Polo": POLO,
        "BT": bt,
        "RR": rr,
        "Sfc_type": sfc_type,
        "Latitude": latitude,
        "Longitude": longitude
    }

    attrs = {"missing_value": -999}
    ds = xr.Dataset(ds_vars, attrs=attrs)
    ds = ds.assign_coords({"Freq": FREQ, "Latitude": latitude, "Longitude": longitude})
    return ds


def _get_datasets_with_less_attributes():
    """Represent files with two resolution of variables in them (ex. OCEAN)."""
    bt = xr.DataArray(np.linspace(1830, 3930, N_SCANLINE * N_FOV * N_CHANNEL, dtype=np.int16).
                      reshape(N_SCANLINE, N_FOV, N_CHANNEL),
                      attrs={"long_name": "Channel Temperature (K)",
                             "scale_factor": 0.01},
                      dims=("Scanline", "Field_of_view", "Channel"))
    rr = xr.DataArray(RANDOM_GEN.integers(100, 500, size=(N_SCANLINE, N_FOV), dtype=np.int16),
                      attrs={"long_name": "Rain Rate (mm/hr)",
                             "scale_factor": 0.1},
                      dims=("Scanline", "Field_of_view"))

    sfc_type = xr.DataArray(RANDOM_GEN.integers(0, 4, size=(N_SCANLINE, N_FOV), dtype=np.int16),
                            attrs={"description": "type of surface:0-ocean," +
                                                  "1-sea ice,2-land,3-snow"},
                            dims=("Scanline", "Field_of_view"))
    latitude = xr.DataArray(DEFAULT_LAT.reshape(DEFAULT_2D_SHAPE),
                            attrs={"long_name":
                                   "Latitude of the view (-90,90)"},
                            dims=("Scanline", "Field_of_view"))
    longitude = xr.DataArray(DEFAULT_LON.reshape(DEFAULT_2D_SHAPE),
                             attrs={"long_name":
                                    "Longitude of the view (-180,180)"},
                             dims=("Scanline", "Field_of_view"))

    ds_vars = {
        "Freq": FREQ,
        "Polo": POLO,
        "BT": bt,
        "RR": rr,
        "Sfc_type": sfc_type,
        "Longitude": longitude,
        "Latitude": latitude
    }

    attrs = {"missing_value": -999.}
    ds = xr.Dataset(ds_vars, attrs=attrs)
    ds = ds.assign_coords({"Freq": FREQ, "Latitude": latitude, "Longitude": longitude})
    return ds


def fake_open_dataset(filename, **kwargs):
    """Create a Dataset similar to reading an actual file with xarray.open_dataset."""
    if filename == METOP_FILE:
        return _get_datasets_with_less_attributes()
    return _get_datasets_with_attributes()


@pytest.mark.parametrize(
    ("filenames", "expected_datasets"),
    [
        ([METOP_FILE], DS_IDS),
        ([NPP_MIRS_L2_SWATH], DS_IDS),
        ([OTHER_MIRS_L2_SWATH], DS_IDS),
    ]
)
def test_available_datasets(filenames, expected_datasets):
    """Test that variables are dynamically discovered."""
    r = _create_fake_reader(filenames, {})
    avails = list(r.available_dataset_names)
    for var_name in expected_datasets:
        assert var_name in avails


@pytest.mark.parametrize(
    ("filenames", "loadable_ids", "platform_name"),
    [
        ([METOP_FILE], TEST_VARS, "metop-a"),
        ([NPP_MIRS_L2_SWATH], TEST_VARS, "npp"),
        ([N20_MIRS_L2_SWATH], TEST_VARS, "noaa-20"),
        ([N21_MIRS_L2_SWATH], TEST_VARS, "noaa-21"),
        ([OTHER_MIRS_L2_SWATH], TEST_VARS, "gpm"),
    ]
)
@pytest.mark.parametrize("reader_kw", [{}, {"limb_correction": False}])
def test_basic_load(filenames, loadable_ids, platform_name, reader_kw):
    """Test that variables are loaded properly."""
    r = _create_fake_reader(filenames, reader_kw)

    test_data = fake_open_dataset(filenames[0])
    exp_limb_corr = reader_kw.get("limb_correction", True) and platform_name in ("npp", "noaa-20", "noaa-21")
    loaded_data_arrs = _load_and_check_limb_correction_variables(r, loadable_ids, platform_name, exp_limb_corr)
    for _data_id, data_arr_dask in loaded_data_arrs.items():
        data_arr = data_arr_dask.compute()
        assert data_arr.dtype == data_arr_dask.dtype
        if np.issubdtype(data_arr.dtype, np.floating):
            # we started with float32, it should stay that way
            # NOTE: Sfc_type does not have enough metadata to dynamically force integer type
            #   even though it is a mask/category product
            assert data_arr.dtype.type == np.float32
        _check_metadata(data_arr, test_data, platform_name)


def _create_fake_reader(
        filenames: list[str],
        reader_kwargs: dict,
        exp_loadable_files: int | None = None
) -> FileYAMLReader:
    exp_loadable_files = exp_loadable_files if exp_loadable_files is not None else len(filenames)
    reader_configs = config_search_paths(os.path.join("readers", "mirs.yaml"))
    with mock.patch("satpy.readers.mirs.xr.open_dataset") as od:
        od.side_effect = fake_open_dataset
        r = load_reader(reader_configs)
        loadables = r.select_files_from_pathnames(filenames)
        r.create_filehandlers(loadables, fh_kwargs=reader_kwargs)

        assert isinstance(r, FileYAMLReader)
        assert len(loadables) == exp_loadable_files
        assert r.file_handlers
    return r


def _load_and_check_limb_correction_variables(
        reader: FileYAMLReader,
        loadable_ids: list[str],
        platform_name: str,
        exp_limb_corr: bool
) -> dict[DataID, xr.DataArray]:
    with mock.patch("satpy.readers.mirs.read_atms_coeff_to_string") as \
            fd, mock.patch("satpy.readers.mirs.retrieve") as rtv:
        fd.side_effect = fake_coeff_from_fn
        loaded_data_arrs = reader.load(loadable_ids)
    if exp_limb_corr:
        fd.assert_called()
        suffix = f"noaa{platform_name[-2:]}" if platform_name.startswith("noaa") else "snpp"
        assert rtv.call_count == 2 * len([var_name for var_name in loadable_ids if "btemp" in var_name])
        for calls_args in rtv.call_args_list:
            assert calls_args[0][0].endswith(f"_{suffix}.txt")
    else:
        fd.assert_not_called()
        rtv.assert_not_called()
    assert len(loaded_data_arrs) == len(loadable_ids)
    return loaded_data_arrs


def _check_metadata(data_arr: xr.DataArray, test_data: xr.Dataset, platform_name: str) -> None:
    var_name = data_arr.attrs["name"]
    if var_name not in ["latitude", "longitude"]:
        _check_area(data_arr)
    assert "_FillValue" not in data_arr.attrs
    _check_attrs(data_arr, platform_name)

    input_fake_data = test_data["BT"] if "btemp" in var_name else test_data[var_name]
    if "valid_range" in input_fake_data.attrs:
        valid_range = input_fake_data.attrs["valid_range"]
        _check_valid_range(data_arr, valid_range)
    if "_FillValue" in input_fake_data.attrs:
        fill_value = input_fake_data.attrs["_FillValue"]
        _check_fill_value(data_arr, fill_value)

    assert data_arr.attrs["units"] == DEFAULT_UNITS[var_name]


def _check_area(data_arr):
    from pyresample.geometry import SwathDefinition
    area = data_arr.attrs["area"]
    assert isinstance(area, SwathDefinition)


def _check_valid_range(data_arr, test_valid_range):
    # valid_range is popped out of data_arr.attrs when it is applied
    assert "valid_range" not in data_arr.attrs
    assert data_arr.data.min() >= test_valid_range[0]
    assert data_arr.data.max() <= test_valid_range[1]


def _check_fill_value(data_arr, test_fill_value):
    assert "_FillValue" not in data_arr.attrs
    assert not (data_arr.data == test_fill_value).any()


def _check_attrs(data_arr, platform_name):
    attrs = data_arr.attrs
    assert "scale_factor" not in attrs
    assert "platform_name" in attrs
    assert attrs["platform_name"] == platform_name
    assert attrs["start_time"] == START_TIME
    assert attrs["end_time"] == END_TIME



# Copyright (c) 2017-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit testing for enhancement color mapping functions."""

import contextlib
import os
from tempfile import NamedTemporaryFile
from unittest import mock

import numpy as np
import pytest

from satpy.enhancements.colormap import create_colormap

from .utils import create_ch1, create_ch2, create_rgb, run_and_check_enhancement


class TestEnhancementsConvolution:
    """Class for testing enhancements in satpy.enhancements.convolution module."""

    def setup_method(self):
        """Create test data used by every test."""
        self.ch1 = create_ch1()
        self.ch2 = create_ch2()
        self.rgb = create_rgb()

    def test_lookup(self):
        """Test the lookup enhancement function."""
        from satpy.enhancements.colormap import lookup
        expected = np.array([[
            [0., 0., 0., 0.333333, 0.705882],
            [1., 1., 1., 1., 1.]]])
        lut = np.arange(256.)
        run_and_check_enhancement(lookup, self.ch1, expected, luts=lut)

        expected = np.array([[[0., 0., 0., 0.333333, 0.705882],
                              [1., 1., 1., 1., 1.]],
                             [[0., 0., 0., 0.333333, 0.705882],
                              [1., 1., 1., 1., 1.]],
                             [[0., 0., 0., 0.333333, 0.705882],
                              [1., 1., 1., 1., 1.]]])
        lut = np.arange(256.)
        lut = np.vstack((lut, lut, lut)).T
        run_and_check_enhancement(lookup, self.rgb, expected, luts=lut)

    def test_colorize(self):
        """Test the colorize enhancement function."""
        from trollimage.colormap import brbg

        from satpy.enhancements.colormap import colorize
        expected = np.array([
            [[np.nan, 3.29411723e-01, 3.29411723e-01, 3.21825881e-08, 3.21825881e-08],
             [3.21825881e-08, 3.21825881e-08, 3.21825881e-08, 3.21825881e-08, 3.21825881e-08]],
            [[np.nan, 1.88235327e-01, 1.88235327e-01, 2.35294109e-01, 2.35294109e-01],
             [2.35294109e-01, 2.35294109e-01, 2.35294109e-01, 2.35294109e-01, 2.35294109e-01]],
            [[np.nan, 1.96078164e-02, 1.96078164e-02, 1.88235281e-01, 1.88235281e-01],
             [1.88235281e-01, 1.88235281e-01, 1.88235281e-01, 1.88235281e-01, 1.88235281e-01]]])
        run_and_check_enhancement(colorize, self.ch1, expected, palettes=brbg)

    def test_palettize(self):
        """Test the palettize enhancement function."""
        from trollimage.colormap import brbg

        from satpy.enhancements.colormap import palettize
        expected = np.array([[[10, 0, 0, 10, 10], [10, 10, 10, 10, 10]]])
        run_and_check_enhancement(palettize, self.ch1, expected, palettes=brbg)

    def test_merge_colormaps(self):
        """Test merging colormaps."""
        from trollimage.colormap import Colormap

        from satpy.enhancements.colormap import _merge_colormaps as mcp
        from satpy.enhancements.colormap import create_colormap
        ret_map = mock.MagicMock()

        create_colormap_mock = mock.Mock(wraps=create_colormap)
        cmap1 = Colormap((1, (1., 1., 1.)))
        kwargs = {"palettes": cmap1}

        with mock.patch("satpy.enhancements.colormap.create_colormap", create_colormap_mock):
            res = mcp(kwargs)
        assert res is cmap1
        create_colormap_mock.assert_not_called()
        create_colormap_mock.reset_mock()
        ret_map.reset_mock()

        cmap1 = {"colors": "blues", "min_value": 0,
                 "max_value": 1}
        kwargs = {"palettes": [cmap1]}
        with mock.patch("satpy.enhancements.colormap.create_colormap", create_colormap_mock), \
                mock.patch("trollimage.colormap.blues", ret_map):
            _ = mcp(kwargs)
        create_colormap_mock.assert_called_once()
        ret_map.reverse.assert_not_called()
        ret_map.set_range.assert_called_with(0, 1)
        create_colormap_mock.reset_mock()
        ret_map.reset_mock()

        cmap2 = {"colors": "blues", "min_value": 2,
                 "max_value": 3, "reverse": True}
        kwargs = {"palettes": [cmap2]}
        with mock.patch("trollimage.colormap.blues", ret_map):
            _ = mcp(kwargs)
        ret_map.reverse.assert_called_once()
        ret_map.set_range.assert_called_with(2, 3)
        create_colormap_mock.reset_mock()
        ret_map.reset_mock()

        kwargs = {"palettes": [cmap1, cmap2]}
        with mock.patch("trollimage.colormap.blues", ret_map):
            _ = mcp(kwargs)
        ret_map.__add__.assert_called_once()

    def tearDown(self):
        """Clean up."""


@contextlib.contextmanager
def closed_named_temp_file(**kwargs):
    """Named temporary file context manager that closes the file after creation.

    This helps with Windows systems which can get upset with opening or
    deleting a file that is already open.

    """
    try:
        with NamedTemporaryFile(delete=False, **kwargs) as tmp_cmap:
            yield tmp_cmap.name
    finally:
        os.remove(tmp_cmap.name)


def _write_cmap_to_file(cmap_filename, cmap_data):
    ext = os.path.splitext(cmap_filename)[1]
    if ext in (".npy",):
        np.save(cmap_filename, cmap_data)
    elif ext in (".npz",):
        np.savez(cmap_filename, cmap_data)
    else:
        np.savetxt(cmap_filename, cmap_data, delimiter=",")


def _generate_cmap_test_data(color_scale, colormap_mode):
    cmap_data = np.array([
        [1, 0, 0],
        [1, 1, 0],
        [1, 1, 1],
        [0, 0, 1],
    ], dtype=np.float64)
    if len(colormap_mode) != 3:
        _cmap_data = cmap_data
        cmap_data = np.empty((cmap_data.shape[0], len(colormap_mode)),
                             dtype=np.float64)
        if colormap_mode.startswith("V") or colormap_mode.endswith("A"):
            cmap_data[:, 0] = np.array([128, 130, 132, 134]) / 255.0
            cmap_data[:, -3:] = _cmap_data
        if colormap_mode.startswith("V") and colormap_mode.endswith("A"):
            cmap_data[:, 1] = np.array([128, 130, 132, 134]) / 255.0
    if color_scale is None or color_scale == 255:
        cmap_data = (cmap_data * 255).astype(np.uint8)
    return cmap_data


class TestColormapLoading:
    """Test utilities used with colormaps."""

    @pytest.mark.parametrize("color_scale", [None, 1.0])
    @pytest.mark.parametrize("colormap_mode", ["RGB", "VRGB", "VRGBA"])
    @pytest.mark.parametrize("extra_kwargs",
                             [
                                 {},
                                 {"min_value": 50, "max_value": 100},
                             ])
    @pytest.mark.parametrize("filename_suffix", [".npy", ".npz", ".csv"])
    def test_cmap_from_file(self, color_scale, colormap_mode, extra_kwargs, filename_suffix):
        """Test that colormaps can be loaded from a binary file."""
        # create the colormap file on disk
        with closed_named_temp_file(suffix=filename_suffix) as cmap_filename:
            cmap_data = _generate_cmap_test_data(color_scale, colormap_mode)
            _write_cmap_to_file(cmap_filename, cmap_data)

            unset_first_value = 128.0 / 255.0 if colormap_mode.startswith("V") else 0.0
            unset_last_value = 134.0 / 255.0 if colormap_mode.startswith("V") else 1.0
            if (color_scale is None or color_scale == 255) and colormap_mode.startswith("V"):
                unset_first_value *= 255
                unset_last_value *= 255
            if "min_value" in extra_kwargs:
                unset_first_value = extra_kwargs["min_value"]
                unset_last_value = extra_kwargs["max_value"]

            first_color = [1.0, 0.0, 0.0]
            if colormap_mode == "VRGBA":
                first_color = [128.0 / 255.0] + first_color

            kwargs1 = {"filename": cmap_filename}
            kwargs1.update(extra_kwargs)
            if color_scale is not None:
                kwargs1["color_scale"] = color_scale

            cmap = create_colormap(kwargs1)

        assert cmap.colors.shape[0] == 4
        np.testing.assert_equal(cmap.colors[0], first_color)
        assert cmap.values.shape[0] == 4
        assert cmap.values[0] == unset_first_value
        assert cmap.values[-1] == unset_last_value

    def test_cmap_vrgb_as_rgba(self):
        """Test that data created as VRGB still reads as RGBA."""
        with closed_named_temp_file(suffix=".npy") as cmap_filename:
            cmap_data = _generate_cmap_test_data(None, "VRGB")
            np.save(cmap_filename, cmap_data)
            cmap = create_colormap({"filename": cmap_filename, "colormap_mode": "RGBA"})

        assert cmap.colors.shape[0] == 4
        assert cmap.colors.shape[1] == 4  # RGBA
        np.testing.assert_equal(cmap.colors[0], [128 / 255., 1.0, 0, 0])
        assert cmap.values.shape[0] == 4
        assert cmap.values[0] == 0
        assert cmap.values[-1] == 1.0

    def test_cmap_with_alpha_set(self):
        """Test that the min_alpha and max_alpha arguments set the alpha channel correctly."""
        with closed_named_temp_file(suffix=".npy") as cmap_filename:
            cmap_data = _generate_cmap_test_data(None, "RGB")
            np.save(cmap_filename, cmap_data)
            cmap = create_colormap({"filename": cmap_filename, "min_alpha": 100, "max_alpha": 255})

        assert cmap.colors.shape[0] == 4
        assert cmap.colors.shape[1] == 4  # RGBA
        # check that we start from min_alpha
        np.testing.assert_equal(cmap.colors[0], [1.0, 0, 0, 100/255.])
        # two thirds of the linear scale
        np.testing.assert_almost_equal(cmap.colors[2], [1., 1., 1., (100+(2/3)*(255-100))/255])
        # check that we end at max_alpha
        np.testing.assert_equal(cmap.colors[3], [0, 0, 1., 1.0])
        # check that values have not been changed
        assert cmap.values.shape[0] == 4
        assert cmap.values[0] == 0
        assert cmap.values[-1] == 1.0

    @pytest.mark.parametrize("alpha_arg", ["max_alpha", "min_alpha"])
    def test_cmap_error_with_only_one_alpha_set(self, alpha_arg):
        """Test that when only min_alpha or max_alpha arguments are set an error is raised."""
        with closed_named_temp_file(suffix=".npy") as cmap_filename:
            cmap_data = _generate_cmap_test_data(None, "RGB")
            np.save(cmap_filename, cmap_data)

            # check that if a value is missing we raise a ValueError
            with pytest.raises(ValueError, match="Both 'min_alpha' and 'max_alpha' must be specified*."):
                create_colormap({"filename": cmap_filename, alpha_arg: 255})

    @pytest.mark.parametrize(
        ("real_mode", "forced_mode"),
        [
            ("VRGBA", "RGBA"),
            ("VRGBA", "VRGB"),
            ("RGBA", "RGB"),
        ]
    )
    @pytest.mark.parametrize("filename_suffix", [".npy", ".csv"])
    def test_cmap_bad_mode(self, real_mode, forced_mode, filename_suffix):
        """Test that reading colormaps with the wrong mode fails."""
        with closed_named_temp_file(suffix=filename_suffix) as cmap_filename:
            cmap_data = _generate_cmap_test_data(None, real_mode)
            _write_cmap_to_file(cmap_filename, cmap_data)
            # Force colormap_mode VRGBA to RGBA and we should see an exception
            with pytest.raises(ValueError, match="Unexpected colormap shape for mode .*"):
                create_colormap({"filename": cmap_filename, "colormap_mode": forced_mode})

    def test_cmap_from_file_bad_shape(self):
        """Test that unknown array shape causes an error."""
        from satpy.enhancements.colormap import create_colormap

        # create the colormap file on disk
        with closed_named_temp_file(suffix=".npy") as cmap_filename:
            np.save(cmap_filename, np.array([
                [0],
                [64],
                [128],
                [255],
            ]))

            with pytest.raises(ValueError, match="Unexpected colormap shape for mode 'None'"):
                create_colormap({"filename": cmap_filename})

    def test_cmap_from_config_path(self, tmp_path):
        """Test loading a colormap relative to a config path."""
        import satpy
        from satpy.enhancements.colormap import create_colormap

        cmap_dir = tmp_path / "colormaps"
        cmap_dir.mkdir()
        cmap_filename = cmap_dir / "my_colormap.npy"
        cmap_data = _generate_cmap_test_data(None, "RGBA")
        np.save(cmap_filename, cmap_data)
        with satpy.config.set(config_path=[tmp_path]):
            rel_cmap_filename = os.path.join("colormaps", "my_colormap.npy")
            cmap = create_colormap({"filename": rel_cmap_filename, "colormap_mode": "RGBA"})

        assert cmap.colors.shape[0] == 4
        assert cmap.colors.shape[1] == 4  # RGBA
        np.testing.assert_equal(cmap.colors[0], [128 / 255., 1.0, 0, 0])
        assert cmap.values.shape[0] == 4
        assert cmap.values[0] == 0
        assert cmap.values[-1] == 1.0

    def test_cmap_from_trollimage(self):
        """Test that colormaps in trollimage can be loaded."""
        from satpy.enhancements.colormap import create_colormap
        cmap = create_colormap({"colors": "pubu"})
        from trollimage.colormap import pubu
        np.testing.assert_equal(cmap.colors, pubu.colors)
        np.testing.assert_equal(cmap.values, pubu.values)

    def test_cmap_no_colormap(self):
        """Test that being unable to create a colormap raises an error."""
        from satpy.enhancements.colormap import create_colormap
        with pytest.raises(ValueError, match="Unknown colormap format: .*"):
            create_colormap({})

    def test_cmap_list(self):
        """Test that colors can be a list/tuple."""
        from satpy.enhancements.colormap import create_colormap
        colors = [
            [0., 0., 1.],
            [1., 0., 1.],
            [0., 1., 1.],
            [1., 1., 1.],
        ]
        values = [2, 4, 6, 8]
        cmap = create_colormap({"colors": colors, "color_scale": 1})
        assert cmap.colors.shape[0] == 4
        np.testing.assert_equal(cmap.colors[0], [0.0, 0.0, 1.0])
        assert cmap.values.shape[0] == 4
        assert cmap.values[0] == 0
        assert cmap.values[-1] == 1.0

        cmap = create_colormap({"colors": colors, "color_scale": 1, "values": values})
        assert cmap.colors.shape[0] == 4
        np.testing.assert_equal(cmap.colors[0], [0.0, 0.0, 1.0])
        assert cmap.values.shape[0] == 4
        assert cmap.values[0] == 2
        assert cmap.values[-1] == 8
